#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.04.29 07:00:00                  #
# ================================================== #

import copy
import os
from pathlib import Path
from uuid import uuid4

from PySide6.QtGui import QAction

from pygpt_net.utils import trans


class Profile:
    def __init__(self, window=None):
        """
        Profile controller

        :param window: Window instance
        """
        self.window = window
        self.dialog = False
        self.width = 500
        self.height = 500
        self.initialized = False
        self.dialog_initialized = False

    def setup(self):
        """Setup profile"""
        self.setup_menu()
        if not self.dialog_initialized:
            self.window.profiles.setup()  # widget dialog
            self.dialog_initialized = True

    def switch(self, uuid: str, force: bool = False, save_current: bool = True):
        """
        Switch profile

        :param uuid: Profile UUID
        :param force: Force switch
        :param save_current: Save current profile
        """
        current = self.window.core.config.profile.get_current()
        if uuid == current and not force:
            self.update_menu()
            return
        profile = self.window.core.config.profile.get(uuid)
        if profile is None:
            self.window.ui.dialogs.alert("Profile not found!")
            return
        self.window.ui.status("Please wait...")

        if save_current:
            print("Saving all settings in current profile...")
            self.window.controller.settings.save_all(force=True)  # save all current settings
            # self.window.controller.layout.save()  # save layout state
        self.window.core.config.profile.set_current(uuid)

        # switch to profile workdir
        path = self.window.core.config.profile.get_current_workdir()
        if path and os.path.exists(path):
            self.window.controller.settings.workdir.update(path, force=True)

        self.update_menu()
        self.update_list()
        self.window.ui.update_title()
        self.window.ui.status(trans("dialog.profile.status.changed") + ": " + profile['name'])
        self.select_current_on_list()

    def select_current_on_list(self):
        """Select current profile on list"""
        current = self.window.core.config.profile.get_current()
        profiles = self.get_profiles()
        idx = 0
        for uuid in profiles:
            if uuid == current:
                index = self.window.ui.models['profile.list'].index(idx, 0)
                self.window.ui.nodes['profile.list'].setCurrentIndex(index)
                break
            idx += 1

    def get_profiles(self) -> dict:
        """
        Get profiles

        :return: profiles dict
        """
        return self.window.core.config.profile.get_all()

    def new(self):
        """New profile dialog"""
        self.window.ui.nodes['dialog.profile.checkbox.switch'].setVisible(True)
        self.window.ui.dialog['profile.item'].checkboxes.setVisible(False)
        self.window.ui.dialog['profile.item'].id = 'profile'
        self.window.ui.dialog['profile.item'].uuid = None
        self.window.ui.dialog['profile.item'].mode = 'create'
        self.window.ui.dialog['profile.item'].path = ""
        self.window.ui.dialog['profile.item'].input.setText("")
        self.window.ui.dialog['profile.item'].prepare()
        self.window.ui.dialog['profile.item'].show()

    def edit(self, uuid: str):
        """
        Edit profile dialog

        :param uuid: profile UUID
        """
        self.window.ui.nodes['dialog.profile.checkbox.switch'].setVisible(False)
        profile = self.window.core.config.profile.get(uuid)
        self.window.ui.dialog['profile.item'].checkboxes.setVisible(False)
        if profile is None:
            self.window.ui.dialogs.alert("Profile not found!")
            return
        self.window.ui.dialog['profile.item'].id = 'profile'
        self.window.ui.dialog['profile.item'].uuid = uuid
        self.window.ui.dialog['profile.item'].mode = 'edit'
        self.window.ui.dialog['profile.item'].path = profile['workdir'].replace("%HOME%", str(Path.home()))
        self.window.ui.dialog['profile.item'].input.setText(profile['name'])
        self.window.ui.dialog['profile.item'].prepare()
        self.window.ui.dialog['profile.item'].show()

    def open(self, force: bool = False):
        """
        Open profiles editor

        :param force: force open
        """
        if not self.initialized:
            self.setup()
            self.initialized = True
        if not self.dialog or force:
            self.window.ui.dialogs.open(
                'profile.editor',
                width=self.width,
                height=self.height,
            )
            self.dialog = True
            self.select_current_on_list()

    def close(self):
        """Close profile dialog"""
        if self.dialog:
            self.window.ui.dialogs.close('profile.editor')
            self.dialog = False

    def toggle_editor(self):
        """Toggle profile dialog"""
        if self.dialog:
            self.close()
        else:
            self.open()

    def handle_update(self, mode: str, name: str, path: str, uuid: str = None):
        """
        Handle new/edit profile dialog

        :param mode: mode (create | update | duplicate)
        :param name: profile name
        :param path: profile workdir path
        :param uuid: profile UUID (update and duplicate only)
        """
        current = self.window.core.config.profile.get_current()
        if name.strip() == "":
            self.window.ui.dialogs.alert(trans("dialog.profile.alert.name.empty"))
            return
        if path.strip() == "":
            self.window.ui.dialogs.alert(trans("dialog.profile.alert.path.empty"))
            return
        if not os.path.exists(path) or not os.path.isdir(path):
            self.window.ui.dialogs.alert(trans("dialog.profile.alert.path.not_exists"))
            return

        if not self.window.core.filesystem.is_directory_empty(path):
            if not self.window.core.filesystem.is_workdir_in_path(path):
                self.window.ui.dialogs.alert(trans("dialog.profile.alert.duplicate.not_empty"))
                return

        if mode == 'create':
            # create new profile
            uuid = self.window.core.config.profile.add(name, path)
            self.window.ui.status(trans("dialog.profile.status.created"))
            if self.window.ui.nodes['dialog.profile.checkbox.switch'].isChecked():
                self.switch(uuid, force=True)

        elif mode == 'edit':
            # update profile
            profile = self.window.core.config.profile.get(uuid)
            old_path = profile['workdir'].replace("%HOME%", str(Path.home()))
            self.window.core.config.profile.update_profile(uuid, name, path)
            self.window.ui.status(trans("dialog.profile.status.updated"))

            # if current profile and path was changed then reload:
            if uuid == current:
                self.window.ui.update_title()
                if old_path != path:
                    self.switch(uuid, force=True)

        elif mode == 'duplicate':
            # duplicate profile (duplicate requires empty directory)
            if not self.window.core.filesystem.is_directory_empty(path):
                self.window.ui.dialogs.alert(trans("dialog.workdir.change.empty.alert"))
                return

            profiles = self.get_profiles()
            if uuid not in profiles:
                self.window.ui.dialogs.alert(trans("dialog.profile.alert.src.empty"))
                return
            profile = profiles[uuid]

            # check if not same path
            if profile['workdir'].replace("%HOME%", str(Path.home())) == path:
                self.window.ui.dialogs.alert(trans("dialog.profile.alert.path.same"))
                return

            # check free space
            include_datadir = self.is_include_datadir()
            include_db = self.is_include_db()
            src_path = profile['workdir'].replace("%HOME%", str(Path.home()))
            space_required = self.window.core.filesystem.get_directory_size(src_path, human_readable=False)
            if not include_datadir:
                space_required -= self.window.core.filesystem.get_datadir_size(src_path, human_readable=False)
            if not include_db:
                space_required -= self.window.core.filesystem.get_db_size(src_path, human_readable=False)
            space_free = self.window.core.filesystem.get_free_disk_space(path, human_readable=False)
            if space_required > space_free:
                self.window.ui.dialogs.alert(trans("dialog.workdir.result.no_free_space").format(
                    required=self.window.core.filesystem.sizeof_fmt(space_required),
                    free=self.window.core.filesystem.sizeof_fmt(space_free),
                ))
                return

            # make duplicate
            self.duplicate(uuid, name, path)
            self.window.ui.status(trans("dialog.profile.status.duplicated"))
            if self.window.ui.nodes['dialog.profile.checkbox.switch'].isChecked():
                self.switch(uuid, force=True)

        # close dialog and update list
        self.window.ui.dialogs.close('profile.item')
        self.update_menu()
        self.update_list()

    def dismiss_update(self):
        """Dismiss update dialog"""
        self.window.ui.dialogs.close('profile.item')

    def edit_by_idx(self, idx: int):
        """
        Rename profile by index

        :param idx: profile index
        """
        uuid = self.get_id_by_idx(idx)
        self.edit(uuid)

    def select_by_idx(self, idx: int):
        """
        Select profile by index

        :param idx: profile index
        """
        uuid = self.get_id_by_idx(idx)
        self.switch(uuid)

    def delete_by_idx(self, idx: int, force: bool = False):
        """
        Delete profile by index

        :param idx: profile index
        :param force: force delete
        """
        uuid = self.get_id_by_idx(idx)
        current = self.window.core.config.profile.get_current()
        if uuid == current:
            self.window.ui.dialogs.alert(trans("dialog.profile.alert.delete.current"))
            return
        if not force:
            self.window.ui.dialogs.confirm(
                type='profile.delete',
                id=idx,
                msg=trans('confirm.profile.delete'),
            )
            return
        self.delete(uuid)

    def delete(self, uuid: str):
        """
        Delete profile (remove only)

        :param uuid: profile ID
        """
        profiles = self.get_profiles()
        if uuid in profiles:
            profile = profiles[uuid]
            name = profile['name']
            # remove profile
            if self.window.core.config.profile.remove(uuid):
                self.window.ui.status(trans("dialog.profile.status.removed") + ": " + name)
                self.update_list()
                self.update_menu()

    def delete_all_by_idx(self, idx: int, force: bool = False):
        """
        Delete profile with files by index

        :param idx: profile index
        :param force: force delete
        """
        uuid = self.get_id_by_idx(idx)
        current = self.window.core.config.profile.get_current()
        if uuid == current:
            self.window.ui.dialogs.alert(trans("dialog.profile.alert.delete.current"))
            return
        if not force:
            self.window.ui.dialogs.confirm(
                type='profile.delete.all',
                id=idx,
                msg=trans('confirm.profile.delete_all'),
            )
            return
        self.delete_all(uuid)

    def delete_all(self, uuid: str):
        """
        Delete profile with files

        :param uuid: profile ID
        """
        profiles = self.get_profiles()
        remove_datadir = True
        remove_db = True
        if uuid in profiles:
            profile = profiles[uuid]
            name = profile['name']
            path = profile['workdir'].replace("%HOME%", str(Path.home()))
            # remove profile
            if self.window.core.config.profile.remove(uuid):
                if not os.path.exists(path) or not os.path.isdir(path):
                    self.window.ui.dialogs.alert(trans("dialog.profile.alert.path.not_exists"))
                    return
                print("Clearing workdir: ", path)
                self.window.core.filesystem.clear_workdir(
                    path,
                    remove_db=remove_db,
                    remove_datadir=remove_datadir,
                )
                self.window.ui.status(trans("dialog.profile.status.deleted") + ": " + name)
                self.update_list()
                self.update_menu()

    def duplicate(self, uuid: str, new_name: str, new_path: str):
        """
        Duplicate profile

        :param uuid: profile ID
        :param new_name: new profile name
        :param new_path: new profile path
        """
        profiles = self.get_profiles()
        if uuid not in profiles:
            self.window.ui.dialogs.alert("Profile not found!")
            return
        profile = profiles[uuid]

        copy_datadir = self.is_include_datadir()
        copy_db = self.is_include_db()

        # make copy
        duplicate = copy.deepcopy(profile)
        new_uuid = str(uuid4())
        duplicate['name'] = new_name
        duplicate['workdir'] = new_path
        self.window.core.config.profile.append(new_uuid, duplicate)

        # copy files from workdir
        path_from = profile['workdir'].replace("%HOME%", str(Path.home()))
        path_to = new_path
        print("Copying all files from {} to: {}".format(path_from, path_to))
        self.window.ui.status("Copying files...")
        result = self.window.core.filesystem.copy_workdir(
            path_from,
            path_to,
            copy_db=copy_db,
            copy_datadir=copy_datadir,
        )
        if not result:
            self.window.ui.dialogs.alert("Error copying files!")
            self.window.ui.status("Error copying files!")
            return
        print("[OK] All files copied successfully.")
        self.window.ui.status("Files copied.")
        self.update_list()
        self.update_menu()

    def duplicate_by_idx(self, idx: int):
        """
        Duplicate profile by index

        :param idx: profile index
        """
        uuid = self.get_id_by_idx(idx)
        profile = self.window.core.config.profile.get(uuid)
        self.window.ui.dialog['profile.item'].checkboxes.setVisible(True)
        if profile is None:
            self.window.ui.dialogs.alert("Profile not found!")
            return
        self.window.ui.nodes['dialog.profile.checkbox.switch'].setVisible(True)
        self.window.ui.dialog['profile.item'].id = 'profile'
        self.window.ui.dialog['profile.item'].uuid = uuid
        self.window.ui.dialog['profile.item'].mode = 'duplicate'
        self.window.ui.dialog['profile.item'].path = ""
        self.window.ui.dialog['profile.item'].input.setText(profile['name'] + " - copy")
        self.window.ui.dialog['profile.item'].prepare()
        self.window.ui.dialog['profile.item'].show()

    def reset(self, uuid: str):
        """
        Reset profile

        :param uuid: profile ID
        """
        profiles = self.get_profiles()
        current = self.window.core.config.profile.get_current()
        remove_datadir = False
        remove_db = True
        if uuid in profiles:
            profile = profiles[uuid]
            path = profile['workdir'].replace("%HOME%", str(Path.home()))
            if not os.path.exists(path) or not os.path.isdir(path):
                self.window.ui.dialogs.alert("Directory not exists!")
                return
            print("Clearing workdir: ", path)
            self.window.core.db.close()
            self.window.core.filesystem.clear_workdir(
                path,
                remove_db=remove_db,
                remove_datadir=remove_datadir,
            )
            if uuid == current:
                self.switch(uuid, force=True, save_current=False)  # reload current profile
            self.window.ui.status("Profile cleared: " + profile['name'])

    def reset_by_idx(self, idx: int, force: bool = False):
        """
        Reset profile by index

        :param idx: profile index
        :param force: force reset
        """
        uuid = self.get_id_by_idx(idx)
        if not force:
            self.window.ui.dialogs.confirm(
                type='profile.reset',
                id=idx,
                msg=trans('confirm.profile.reset'),
            )
            return
        self.reset(uuid)

    def is_include_db(self):
        """Get include db"""
        return self.window.ui.nodes['dialog.profile.checkbox.db'].isChecked()

    def is_include_datadir(self):
        """Get include datadir"""
        return self.window.ui.nodes['dialog.profile.checkbox.data'].isChecked()

    def get_id_by_idx(self, idx: int) -> str:
        """
        Get profile id by index

        :param idx: profile index
        """
        profiles = self.get_profiles()
        if len(profiles) > idx:
            return list(profiles.keys())[idx]

    def update_list(self):
        """Reload profile list"""
        items = self.get_profiles()
        self.window.profiles.update_list("profile.list", items)

    def setup_menu(self):
        """Setup profile menu"""
        profiles = self.window.core.config.profile.get_all()
        current = self.window.core.config.profile.get_current()
        for uuid in profiles:
            if uuid not in self.window.ui.menu['config.profiles']:
                profile = profiles[uuid]
                name = profile['name']
                checked = False
                if uuid == current:
                    name = name + " " + trans("profile.current.suffix")
                    checked = True
                self.window.ui.menu['config.profiles'][uuid] = QAction(
                    name,
                    self.window,
                    checkable=True,
                )
                self.window.ui.menu['config.profiles'][uuid].setChecked(checked)
                self.window.ui.menu['config.profiles'][uuid].triggered.connect(
                    lambda checked=True, uuid=uuid: self.window.controller.settings.profile.switch(uuid))
                self.window.ui.menu['config.profile'].addAction(self.window.ui.menu['config.profiles'][uuid])

    def update_menu(self):
        """Update menu"""
        profiles = self.window.core.config.profile.get_all()
        current = self.window.core.config.profile.get_current()
        for uuid in list(self.window.ui.menu['config.profiles'].keys()):
            if uuid in profiles:
                name = profiles[uuid]['name']
                checked = False
                if uuid == current:
                    name = name + " " + trans("profile.current.suffix")
                    checked = True
                self.window.ui.menu['config.profiles'][uuid].setText(name)
                self.window.ui.menu['config.profiles'][uuid].setChecked(checked)

        # add new profiles
        for uuid in list(profiles.keys()):
            if uuid not in self.window.ui.menu['config.profiles']:
                profile = profiles[uuid]
                name = profile['name']
                checked = False
                if uuid == current:
                    name = name + " " + trans("profile.current.suffix")
                    checked = True
                self.window.ui.menu['config.profiles'][uuid] = QAction(
                    name,
                    self.window,
                    checkable=True,
                )
                self.window.ui.menu['config.profiles'][uuid].setChecked(checked)
                self.window.ui.menu['config.profiles'][uuid].triggered.connect(
                    lambda checked=True, uuid=uuid: self.window.controller.settings.profile.switch(uuid)
                )
                self.window.ui.menu['config.profile'].addAction(self.window.ui.menu['config.profiles'][uuid])

        # remove non-exist profiles
        for uuid in list(self.window.ui.menu['config.profiles'].keys()):
            if uuid not in profiles:
                self.window.ui.menu['config.profile'].removeAction(self.window.ui.menu['config.profiles'][uuid])
                del self.window.ui.menu['config.profiles'][uuid]




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.04.17 01:00:00                  #
# ================================================== #

from llama_index.core.readers.base import BaseReader

from .hub.video_audio.base import VideoAudioReader
from .base import BaseLoader


class Loader(BaseLoader):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.id = "video_audio"
        self.name = "Video/audio"
        self.extensions = ["mp4", "avi", "mov", "mkv", "webm", "mp3", "mpeg", "mpga", "m4a", "wav"]
        self.type = ["file"]
        self.init_args = {
            "use_local": False,  # use local model instead of API
            "model_version": "base", # Whisper model version
        }
        self.init_args_types = {
            "use_local": "bool",
            "model_version": "str",
        }
        """
        https://github.com/openai/whisper
        Allowed models:
        - tiny (39 M)
        - base (74 M)
        - small (244 M)
        - medium (769 M)
        - large (1550 M)
        - large-v2 (1550 M)
        """

    def get(self) -> BaseReader:
        """
        Get reader instance

        :return: Data reader instance
        """
        args = self.get_args()
        args["window"] = self.window  # pass window instance
        if self.window is not None:
            args["use_local"] = self.window.core.config.get("llama.hub.loaders.use_local", False)
        return VideoAudioReader(**args)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.04.29 16:00:00                  #
# ================================================== #

import time

from sqlalchemy import text

from pygpt_net.item.assistant import AssistantFileItem
from .utils import pack_item_value, unpack_file


class Storage:
    def __init__(self, window=None):
        """
        Initialize storage instance

        :param window: Window instance
        """
        self.window = window

    def attach(self, window):
        """
        Attach window instance

        :param window: Window instance
        """
        self.window = window

    def get_all(self) -> dict:
        """
        Return dict with AssistantFileItem objects, indexed by ID

        :return: dict of AssistantFileItem objects
        """
        stmt = text("""
            SELECT * FROM remote_file
        """)
        items = {}
        db = self.window.core.db.get_db()
        with db.connect() as conn:
            result = conn.execute(stmt)
            for row in result:
                file = AssistantFileItem()
                unpack_file(file, row._asdict())
                items[file.id] = file  # index by ID
        return items

    def get_by_id(self, id: int) -> AssistantFileItem:
        """
        Return AssistantFileItem by ID

        :param id: file item ID
        :return: AssistantFileItem
        """
        stmt = text("""
            SELECT * FROM remote_file WHERE id = :id LIMIT 1
        """).bindparams(idx=id)
        file = AssistantFileItem()
        db = self.window.core.db.get_db()
        with db.connect() as conn:
            result = conn.execute(stmt)
            for row in result:
                unpack_file(file, row._asdict())
        return file

    def get_by_store_or_thread(self, store_id: str, thread_id: str) -> dict:
        """
        Return dict with AssistantFileItem objects, indexed by ID

        :return: dict of AssistantFileItem objects
        """
        if (store_id is None or store_id == "") and (thread_id is None or thread_id == ""):
            return {}

        if store_id is not None and store_id != "" and thread_id is not None and thread_id != "":
            stmt = text("""
                SELECT * FROM remote_file WHERE thread_id = :thread OR store_id = :store
            """).bindparams(
                thread=thread_id,
                store=store_id
            )
        elif store_id is not None and store_id != "":
            stmt = text("""
                SELECT * FROM remote_file WHERE store_id = :store
            """).bindparams(
                store=store_id
            )
        elif thread_id is not None and thread_id != "":
            stmt = text("""
                SELECT * FROM remote_file WHERE thread_id = :thread
            """).bindparams(
                thread=thread_id
            )
        items = {}
        db = self.window.core.db.get_db()
        with db.connect() as conn:
            result = conn.execute(stmt)
            for row in result:
                file = AssistantFileItem()
                unpack_file(file, row._asdict())
                items[file.id] = file  # index by record ID
        return items

    def count_by_store_or_thread(self, store_id: str, thread_id: str) -> int:
        """
        Count AssistantFileItem objects

        :return: num of files
        """
        if (store_id is None or store_id == "") and (thread_id is None or thread_id == ""):
            return 0

        if store_id is not None and store_id != "" and thread_id is not None and thread_id != "":
            stmt = text("""
                SELECT COUNT(*) FROM remote_file WHERE thread_id = :thread OR store_id = :store
            """).bindparams(
                thread=thread_id,
                store=store_id
            )
        elif store_id is not None and store_id != "":
            stmt = text("""
                SELECT COUNT(*) FROM remote_file WHERE store_id = :store
            """).bindparams(
                store=store_id
            )
        elif thread_id is not None and thread_id != "":
            stmt = text("""
                SELECT COUNT(*) FROM remote_file WHERE thread_id = :thread
            """).bindparams(
                thread=thread_id
            )
        num = 0
        db = self.window.core.db.get_db()
        with db.connect() as conn:
            result = conn.execute(stmt)
            for row in result:
                num = row[0]
        return num

    def get_all_by_file_id(self, file_id: str) -> dict:
        """
        Get all files by file ID

        :param file_id: file ID
        :return: files dict
        """
        stmt = text("""
            SELECT * FROM remote_file WHERE file_id = :file_id
        """).bindparams(file_id=file_id)
        items = {}
        db = self.window.core.db.get_db()
        with db.connect() as conn:
            result = conn.execute(stmt)
            for row in result:
                file = AssistantFileItem()
                unpack_file(file, row._asdict())
                items[file.id] = file
        return items

    def truncate_all(self) -> bool:
        """
        Truncate all files items

        :return: True if truncated
        """
        db = self.window.core.db.get_db()
        with db.begin() as conn:
            conn.execute(text("DELETE FROM remote_file"))
            conn.execute(text("DELETE FROM sqlite_sequence WHERE name='remote_file'"))
        return True

    def truncate_by_store(self, store_id: str) -> bool:
        """
        Truncate all files items by store ID

        :param store_id: store ID
        :return: True if truncated
        """
        stmt = text("""
            DELETE FROM remote_file WHERE store_id = :store_id
        """).bindparams(store_id=store_id)
        db = self.window.core.db.get_db()
        with db.begin() as conn:
            conn.execute(stmt)
            return True

    def delete_by_id(self, id: int) -> bool:
        """
        Delete file item by record ID

        :param id: record ID
        :return: True if deleted
        """
        stmt = text("""
            DELETE FROM remote_file WHERE id = :id
        """).bindparams(id=id)
        db = self.window.core.db.get_db()
        with db.begin() as conn:
            conn.execute(stmt)
            return True

    def delete_by_file_id(self, file_id: str) -> bool:
        """
        Delete file item by record ID

        :param file_id: file ID
        :return: True if deleted
        """
        stmt = text("""
            DELETE FROM remote_file WHERE file_id = :id
        """).bindparams(id=file_id)
        db = self.window.core.db.get_db()
        with db.begin() as conn:
            conn.execute(stmt)
            return True

    def clear_store_from_files(self, store_id: str) -> bool:
        """
        Clear store from files

        :param store_id: store ID
        :return: True if cleared
        """
        stmt = text("""
            DELETE FROM remote_file WHERE store_id = :store_id
        """).bindparams(store_id=store_id)
        db = self.window.core.db.get_db()
        with db.begin() as conn:
            conn.execute(stmt)
            return True

    def clear_all_stores_from_files(self) -> bool:
        """
        Clear all stores from files

        :return: True if cleared
        """
        stmt = text("""
            DELETE FROM remote_file WHERE store_id IS NOT NULL
        """)
        db = self.window.core.db.get_db()
        with db.begin() as conn:
            conn.execute(stmt)
            return True

    def rename_file(self, record_id: int, name: str) -> bool:
        """
        Rename file

        :param record_id: record ID
        :param name: new name
        :return: True if renamed
        """
        stmt = text("""
            UPDATE remote_file SET name = :name WHERE id = :id
        """).bindparams(
            id=record_id,
            name=name
        )
        db = self.window.core.db.get_db()
        with db.begin() as conn:
            conn.execute(stmt)
            return True

    def save(self, file: AssistantFileItem):
        """
        Insert or update file item

        :param file: AssistantFileItem object
        """
        db = self.window.core.db.get_db()
        with db.begin() as conn:
            ts = int(time.time())
            stmt = text("""
                UPDATE remote_file
                SET
                    name = :name,
                    path = :path,
                    file_id = :file_id,
                    store_id = :store_id,
                    thread_id = :thread_id,
                    updated_ts = :updated_ts
                WHERE id = :id
                """).bindparams(
                id=file.record_id,
                name=file.name,
                path=file.path,
                file_id=file.file_id,
                store_id=file.store_id,
                thread_id=file.thread_id,
                updated_ts=int(ts or 0)
            )
            conn.execute(stmt)

    def insert(self, file: AssistantFileItem) -> int:
        """
        Insert file item

        :param file: AssistantFileItem object
        :return: file item ID
        """
        db = self.window.core.db.get_db()
        ts = int(time.time())
        stmt = text("""
                INSERT INTO remote_file
                (
                    uuid,
                    name,
                    path,
                    size,
                    file_id,
                    store_id,
                    thread_id,
                    created_ts,
                    updated_ts
                )
                VALUES
                (
                    :uuid,
                    :name,
                    :path,
                    :size,
                    :file_id,
                    :store_id,
                    :thread_id,
                    :created_ts,
                    :updated_ts
                )
            """).bindparams(
            uuid=file.uuid,
            name=file.name,
            path=file.path,
            size=int(file.size or 0),
            file_id=file.file_id,
            store_id=file.store_id,
            thread_id=file.thread_id,
            created_ts=int(ts or 0),
            updated_ts=int(ts or 0)
        )
        with db.begin() as conn:
            result = conn.execute(stmt)
            file.record_id = result.lastrowid
            return file.record_id



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.05.01 17:00:00                  #
# ================================================== #

from pygpt_net.core.dispatcher import Event

from .template import Template
from ...item.ctx import CtxItem


class Prompt:
    def __init__(self, window=None):
        """
        Prompt core

        :param window: Window instance
        """
        self.window = window
        self.template = Template(window)

    def get(self, prompt: str) -> str:
        """
        Get prompt content

        :param prompt: id of the prompt
        :return: text content
        """
        key = "prompt." + prompt
        if self.window.core.config.has(key):
            return str(self.window.core.config.get(key))
        return ""

    def build_final_system_prompt(self, prompt: str) -> str:
        # tmp dispatch event: system prompt
        event = Event(Event.SYSTEM_PROMPT, {
            'mode': self.window.core.config.get('mode'),
            'value': prompt,
            'silent': True,
        })
        self.window.core.dispatcher.dispatch(event)
        prompt = event.data['value']

        if self.window.core.config.get('cmd') or self.window.controller.plugins.is_type_enabled("cmd.inline"):

            # cmd syntax tokens
            data = {
                'prompt': prompt,
                'silent': True,
                'syntax': [],
                'cmd': [],
            }

            # tmp dispatch event: command syntax apply
            # full execute cmd syntax
            if self.window.core.config.get('cmd'):
                event = Event(Event.CMD_SYNTAX, data)
                self.window.core.dispatcher.dispatch(event)
                prompt = self.window.core.command.append_syntax(event.data)

            # inline cmd syntax only
            elif self.window.controller.plugins.is_type_enabled("cmd.inline"):
                event = Event(Event.CMD_SYNTAX_INLINE, data)
                self.window.core.dispatcher.dispatch(event)
                prompt = self.window.core.command.append_syntax(event.data)

        return prompt

    def prepare_sys_prompt(
            self,
            mode: str,
            sys_prompt: str,
            ctx: CtxItem,
            reply: bool,
            internal: bool
    ):
        """
        Prepare system prompt

        :param mode: mode
        :param sys_prompt: system prompt
        :param ctx: context item
        :param reply: reply from plugins
        :param internal: internal call
        :return: system prompt
        """
        # event: system prompt (append to system prompt)
        event = Event(Event.SYSTEM_PROMPT, {
            'mode': mode,
            'value': sys_prompt,
        })
        self.window.core.dispatcher.dispatch(event)
        sys_prompt = event.data['value']

        # event: post prompt (post-handle system prompt)
        event = Event(Event.POST_PROMPT, {
            'mode': mode,
            'reply': reply,
            'internal': internal,
            'value': sys_prompt,
        })
        event.ctx = ctx
        self.window.core.dispatcher.dispatch(event)
        sys_prompt = event.data['value']

        # event: command syntax apply (if commands enabled or inline plugin then append commands prompt)
        if self.window.core.config.get('cmd') or self.window.controller.plugins.is_type_enabled("cmd.inline"):
            data = {
                'mode': mode,
                'prompt': sys_prompt,
                'syntax': [],
                'cmd': [],
            }
            # full execute cmd syntax
            if self.window.core.config.get('cmd'):
                event = Event(Event.CMD_SYNTAX, data)
                self.window.core.dispatcher.dispatch(event)
                sys_prompt = self.window.core.command.append_syntax(event.data)

            # inline cmd syntax only
            elif self.window.controller.plugins.is_type_enabled("cmd.inline"):
                event = Event(Event.CMD_SYNTAX_INLINE, data)
                self.window.core.dispatcher.dispatch(event)
                sys_prompt = self.window.core.command.append_syntax(event.data)

        return sys_prompt



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.04.26 23:00:00                  #
# ================================================== #

import json
import time

from pygpt_net.item.attachment import AttachmentItem


class AssistantItem:
    def __init__(self):
        """Assistant item"""
        self.id = None
        self.name = None
        self.description = None
        self.instructions = None
        self.model = None
        self.meta = {}
        self.files = {}  # file IDs (files uploaded to remote storage)
        self.attachments = {}  # local attachments
        self.vector_store = "" # vector store ID
        self.tools = {
            "code_interpreter": False,
            "file_search": False,
            "function": [],
        }

    def reset(self):
        """Reset assistant"""
        self.id = None
        self.name = None
        self.description = None
        self.instructions = None
        self.model = None
        self.meta = {}
        self.files = {}
        self.attachments = {}
        self.vector_store = ""
        self.tools = {
            "code_interpreter": False,
            "file_search": False,
            "function": [],
        }

    def to_dict(self) -> dict:
        """
        Return as dictionary

        :return: dictionary
        """
        return {
            "id": self.id,
            "name": self.name,
            "description": self.description,
            "instructions": self.instructions,
            "model": self.model,
            "meta": self.meta,
            "files": self.files,
            "attachments": self.attachments,
            "vector_store": self.vector_store,
            "tool.code_interpreter": self.tools["code_interpreter"],
            "tool.file_search": self.tools["file_search"],
            "tool.function": self.tools["function"],
        }

    def clear_functions(self):
        """Clear functions"""
        self.tools['function'] = []

    def clear_tools(self):
        """Clear tools"""
        self.tools = {
            "code_interpreter": False,
            "file_search": False,
            "function": [],
        }

    def add_function(self, name: str, parameters: str, desc: str):
        """
        Add function to assistant

        :param name: function name
        :param parameters: function parameters (JSON encoded)
        :param desc: function description
        """
        function = {
            'name': name,
            'params': parameters,
            'desc': desc,
        }
        self.tools['function'].append(function)

    def has_functions(self) -> bool:
        """
        Check if assistant has functions

        :return: True if has functions
        """
        return len(self.tools['function']) > 0

    def get_functions(self) -> list:
        """
        Return assistant functions

        :return: functions list
        """
        return self.tools['function']

    def has_tool(self, tool: str) -> bool:
        """
        Check if assistant has tool

        :param tool: tool name
        :return: bool
        """
        return tool in self.tools and self.tools[tool] is True

    def has_file(self, id: str) -> bool:
        """
        Check if assistant has file with ID

        :param id: file ID
        :return: True if has file
        """
        return id in self.files

    def add_file(self, id: str):
        """
        Add empty file to assistant

        :param id: file ID
        """
        self.files[id] = {}
        self.files[id]['id'] = id

    def delete_file(self, file_id: str):
        """
        Delete file from assistant

        :param file_id: file ID
        """
        if file_id in self.files:
            self.files.pop(file_id)

    def clear_files(self):
        """Clear files dict"""
        self.files = {}

    def has_attachment(self, id: str) -> bool:
        """
        Check if assistant has attachment with ID

        :param id: attachment ID
        :return: True if has attachment
        """
        return id in self.attachments

    def add_attachment(self, attachment: AttachmentItem):
        """
        Add attachment to assistant

        :param attachment: attachment
        """
        id = attachment.id
        self.attachments[id] = attachment

    def delete_attachment(self, id: str):
        """
        Delete attachment from assistant

        :param id: attachment ID
        """
        if id in self.attachments:
            self.attachments.pop(id)

    def clear_attachments(self):
        """Clear attachments dict"""
        self.attachments = {}

    def dump(self) -> str:
        """
        Dump item to string

        :return: serialized item
        """
        try:
            return json.dumps(self.to_dict())
        except Exception as e:
            pass
        return ""

    def __str__(self):
        """To string"""
        return self.dump()


class AssistantStoreItem:
    def __init__(self):
        """Assistant vector store item"""
        self.id = None
        self.record_id = None
        self.uuid = None
        self.name = None
        self.description = None
        self.status = {}
        self.last_status = ""
        self.expire_days = 0
        self.usage_bytes = 0
        self.bytes = 0
        self.num_files = 0
        self.is_thread = False
        self.created = int(time.time())
        self.updated = int(time.time())
        self.last_active = int(time.time())
        self.last_sync = int(time.time())

    def reset(self):
        """Reset store"""
        self.id = None
        self.record_id = None
        self.uuid = None
        self.name = None
        self.description = None
        self.status = {}
        self.last_status = ""
        self.expire_days = 0
        self.usage_bytes = 0
        self.num_files = 0
        self.bytes = 0
        self.is_thread = False
        self.last_active = int(time.time())
        self.last_sync = int(time.time())

    def to_dict(self) -> dict:
        """
        Return as dictionary

        :return: dictionary
        """
        return {
            "id": self.id,
            "uuid": self.uuid,
            "name": self.name,
            "description": self.description,
            "last_status": self.last_status,
            "expire_days": self.expire_days,
            "usage_bytes": self.usage_bytes,
            "num_files": self.num_files,
            "status": self.status,
            "is_thread": self.is_thread,
        }

    def from_dict(self, data: dict):
        """
        Load from dictionary

        :param data: dictionary
        """
        self.id = data.get('id', None)
        self.name = data.get('name', None)
        self.uuid = data.get('uuid', None)
        self.expire_days = data.get('expire_days', 0)
        self.status = data.get('status', {})

    def dump(self) -> str:
        """
        Dump item to string

        :return: serialized item
        """
        try:
            return json.dumps(self.to_dict())
        except Exception as e:
            pass
        return ""

    def __str__(self):
        """To string"""
        return self.dump()


class AssistantFileItem:
    def __init__(self):
        """Assistant file item"""
        self.id = None
        self.record_id = None
        self.name = None
        self.path = None
        self.file_id = None
        self.store_id = None
        self.thread_id = None
        self.uuid = None
        self.size = 0
        self.created = 0
        self.updated = 0

    def reset(self):
        """Reset store"""
        self.id = None
        self.record_id = None
        self.name = None
        self.path = None
        self.file_id = None
        self.store_id = None
        self.thread_id = None
        self.uuid = None
        self.size = 0
        self.created = 0
        self.updated = 0

    def to_dict(self) -> dict:
        """
        Return as dictionary

        :return: dictionary
        """
        return {
            "id": self.id,
            "name": self.name,
            "path": self.path,
            "file_id": self.file_id,
            "store_id": self.store_id,
            "thread_id": self.thread_id,
            "uuid": self.uuid,
            "size": self.size,
            "created": self.created,
            "updated": self.updated,
        }

    def from_dict(self, data: dict):
        """
        Load from dictionary

        :param data: dictionary
        """
        self.id = data.get('id', None)
        self.name = data.get('name', None)
        self.path = data.get('path', None)
        self.file_id = data.get('file_id', None)
        self.store_id = data.get('store_id', None)
        self.thread_id = data.get('thread_id', None)
        self.uuid = data.get('uuid', None)
        self.size = data.get('size', 0)
        self.created = data.get('created', 0)
        self.updated = data.get('updated', 0)

    def dump(self) -> str:
        """
        Dump item to string

        :return: serialized item
        """
        try:
            return json.dumps(self.to_dict())
        except Exception as e:
            pass
        return ""

    def __str__(self):
        """To string"""
        return self.dump()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.03.01 03:00:00                  #
# ================================================== #

from typing import cast

from .base import BaseProvider


class OpenAIWhisperLocal(BaseProvider):
    def __init__(self, *args, **kwargs):
        """
        OpenAI Whisper provider (local model)

        :param args: args
        :param kwargs: kwargs
        """
        super(OpenAIWhisperLocal, self).__init__(*args, **kwargs)
        self.plugin = kwargs.get("plugin")
        self.id = "openai_whisper_local"
        self.name = "Whisper (local)"

    def init_options(self):
        """Initialize options"""
        self.plugin.add_option(
            "whisper_local_model",
            type="text",
            value="base",
            label="Model",
            tab="openai_whisper_local",
            description="Specify local model, default: base. Local models are not available in compiled version, use "
                        "API version if you are using compiled or Snap version.",
            urls={
                "Models": "https://github.com/openai/whisper"
            }
        )

    def transcribe(self, path: str) -> str:
        """
        Audio to text transcription

        :param path: path to audio file to transcribe
        :return: transcribed text
        """
        is_compiled = self.plugin.window.core.config.is_compiled() or self.plugin.window.core.platforms.is_snap()
        if is_compiled:
            raise ValueError("Local models are not available in compiled version.")

        if not self.is_configured():
            raise ImportError(self.get_config_message())

        import whisper

        model = whisper.load_model(self.plugin.get_option_value('whisper_local_model'))
        config = {"model": model}
        model = cast(whisper.Whisper, config["model"])
        result = model.transcribe(path)
        return str(result["text"])

    def is_configured(self) -> bool:
        """
        Check if provider is configured

        :return: True if configured, False otherwise
        """
        is_compiled = self.plugin.window.core.config.is_compiled() or self.plugin.window.core.platforms.is_snap()
        if is_compiled:
            raise ValueError("Local models are not available in compiled version.")
        try:
            import whisper
        except ImportError:
            return False
        return True

    def get_config_message(self) -> str:
        """
        Return message to display when provider is not configured

        :return: message
        """
        return ("Please install OpenAI whisper model "
                "'pip install git+https://github.com/openai/whisper.git' "
                "or pip install openai-whisper to use the model")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.05.01 17:00:00                  #
# ================================================== #

import os
import shutil

from packaging.version import parse as parse_version, Version


class Patch:
    def __init__(self, window=None):
        self.window = window

    def execute(self, version: Version) -> bool:
        """
        Migrate to current app version

        :param version: current app version
        :return: True if migrated
        """
        migrated = False
        is_llama = False
        is_expert = False
        for k in self.window.core.presets.items:
            data = self.window.core.presets.items[k]
            updated = False

            # get version of preset
            if data.version is None or data.version == "":
                continue

            old = parse_version(data.version)

            # check if presets file is older than current app version
            if old < version:
                # < 2.0.0
                if old < parse_version("2.0.0"):
                    print("Migrating presets dir from < 2.0.0...")
                    self.window.core.updater.patch_file('presets', True)  # force replace file

                # < 2.0.53
                if old < parse_version("2.0.53") and k == 'current.assistant':
                    print("Migrating preset file from < 2.0.53...")
                    dst = os.path.join(self.window.core.config.get_user_dir('presets'), 'current.assistant.json')
                    src = os.path.join(self.window.core.config.get_app_path(), 'data', 'config', 'presets',
                                       'current.assistant.json')
                    shutil.copyfile(src, dst)
                    updated = True
                    print("Patched file: {}.".format(dst))

                # < 2.0.102
                if old < parse_version("2.0.102"):
                    if 'current.llama_index' not in self.window.core.presets.items and not is_llama:
                        print("Migrating preset file from < 2.0.102...")
                        dst = os.path.join(self.window.core.config.get_user_dir('presets'), 'current.llama_index.json')
                        src = os.path.join(self.window.core.config.get_app_path(), 'data', 'config', 'presets',
                                           'current.llama_index.json')
                        shutil.copyfile(src, dst)
                        updated = True
                        is_llama = True  # prevent multiple copies
                        print("Patched file: {}.".format(dst))

                # < 2.2.7
                if old < parse_version("2.2.7"):
                    if not is_expert:
                        print("Migrating preset files from < 2.2.7...")
                        dst = os.path.join(self.window.core.config.get_user_dir('presets'), 'current.expert.json')
                        src = os.path.join(self.window.core.config.get_app_path(), 'data', 'config', 'presets', 'current.expert.json')
                        shutil.copyfile(src, dst)
                        print("Patched file: {}.".format(dst))
                        dst = os.path.join(self.window.core.config.get_user_dir('presets'), 'current.agent.json')
                        src = os.path.join(self.window.core.config.get_app_path(), 'data', 'config', 'presets', 'current.agent.json')
                        shutil.copyfile(src, dst)
                        print("Patched file: {}.".format(dst))
                        dst = os.path.join(self.window.core.config.get_user_dir('presets'), 'joke_expert.json')
                        src = os.path.join(self.window.core.config.get_app_path(), 'data', 'config', 'presets', 'joke_expert.json')
                        shutil.copyfile(src, dst)
                        print("Patched file: {}.".format(dst))
                        updated = True
                        is_expert = True  # prevent multiple copies

            # update file
            if updated:
                self.window.core.presets.load()  # reload presets from patched files
                self.window.core.presets.save(k)  # re-save presets
                migrated = True
                print("Preset {} patched to version {}.".format(k, version))

        return migrated



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.05.02 19:00:00                  #
# ================================================== #


class Experts:
    def __init__(self, window=None):
        """
        Experts controller

        :param window: Window instance
        """
        self.window = window
        self.is_stop = False

    def stop(self):
        """
        Stop experts
        """
        self.is_stop = True

    def stopped(self) -> bool:
        """
        Check if experts are stopped

        :return: True if experts are stopped
        """
        return self.is_stop

    def unlock(self):
        """
        Unlock experts
        """
        self.is_stop = False

    def enabled(self) -> bool:
        """
        Check if experts are enabled

        :return: True if experts are enabled
        """
        modes = ["agent", "expert"]
        mode = self.window.core.config.get('mode')
        if mode in modes or self.window.controller.plugins.is_type_enabled("expert"):
            return True
        return False


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.03.26 15:00:00                  #
# ================================================== #

import os

from PySide6.QtGui import QAction, QIcon
from PySide6.QtWidgets import QFileDialog

from pygpt_net.tools.base import BaseTool
from pygpt_net.tools.audio_transcriber.ui.dialogs import AudioTranscribe
from pygpt_net.core.dispatcher import Event
from pygpt_net.item.ctx import CtxItem
from pygpt_net.utils import trans


class AudioTranscriber(BaseTool):
    def __init__(self, *args, **kwargs):
        """
        Audio Transcriber

        :param window: Window instance
        """
        super(AudioTranscriber, self).__init__(*args, **kwargs)
        self.id = "transcriber"
        self.opened = False
        self.video_extensions = ["mp4", "avi", "mov", "mkv", "webm"]
        self.dialog = None  # dialog

    def setup(self):
        """Setup controller"""
        self.restore()
        self.restore_auto_convert()

    def update(self):
        """Update"""
        pass

    def open_file(self):
        """Open transcribe file dialog"""
        path, _ = QFileDialog.getOpenFileName(
            self.window,
            trans("action.video.open"),
            "",
            "Audio Files (*.mp3 *.wav *.ogg *.flac *.m4a *.mp4 *.avi *.mov *.mkv *.webm)")
        if path:
            self.transcribe(path)

    def save_as_file(self):
        """Save transcription as file"""
        text = self.window.ui.editor["audio.transcribe"].toPlainText()
        self.window.controller.chat.common.save_text(text)

    def from_file(self, path: str):
        """
        Open and transcribe audio file

        :param path: audio file path
        """
        self.open()
        self.window.ui.nodes['audio.transcribe.status'].setText(
            trans("audio.transcribe.result.selected").format(path=os.path.basename(path)))
        self.transcribe(path)

    def open(self):
        """Open transcriber"""
        self.window.ui.nodes['audio.transcribe.status'].setText("")
        self.window.ui.dialogs.open('audio.transcribe', width=800, height=600)
        self.opened = True
        self.update()

    def close(self):
        """Close transcribe"""
        self.window.ui.dialogs.close('audio.transcribe')
        self.opened = False

    def toggle(self):
        """Toggle transcribe window"""
        if self.opened:
            self.close()
        else:
            self.open()

    def transcribe(self, path: str, force: bool = False):
        """
        Transcribe audio file

        :param path: audio file path
        :param force: force transcribe
        """
        if not force:
            self.window.ui.dialogs.confirm(
                type='audio.transcribe',
                id=path,
                msg=trans("audio.transcribe.confirm"),
            )
            return
        path = self.prepare_audio(path)
        if path is not None:
            event = Event(Event.AUDIO_INPUT_TRANSCRIBE, {
                'path': str(path),
            })
            event.ctx = CtxItem()  # tmp
            self.clear(force=True)
            self.window.controller.command.dispatch_only(event)
            self.window.ui.nodes['audio.transcribe.status'].setText(
                "Transcribing: {} ... Please wait...".format(os.path.basename(path)))

    def on_transcribe(self, path: str, text: str):
        """
        On audio transcribed

        :param path: audio file path
        :param text: transcribed text
        """
        self.store(text)
        self.window.ui.nodes['audio.transcribe.status'].setText(
            trans("audio.transcribe.result.finished").format(path=os.path.basename(path)))
        self.window.ui.editor["audio.transcribe"].setPlainText(text)

    def prepare_audio(self, path: str) -> str:
        """
        Convert video to audio if needed

        :param path: path to video/audio file
        """
        # ffmpeg required here
        if self.is_video(path) and self.is_auto_convert():
            try:
                self.window.ui.nodes['audio.transcribe.status'].setText(
                    "Converting to audio: {} ... Please wait...".format(os.path.basename(path)))
                video_type = path.split(".")[-1].lower()
                try:
                    from pydub import AudioSegment
                except ImportError:
                    self.window.ui.nodes['audio.transcribe.status'].setText("Please install pydub 'pip install pydub'")
                    raise ImportError("Please install pydub 'pip install pydub' ")
                video = AudioSegment.from_file(path, format=video_type)
                # extract audio from video
                audio = video.split_to_mono()[0]
                file_str = os.path.join(self.window.core.config.get_user_path(), "transcript.mp3")
                if os.path.exists(file_str):
                    os.remove(file_str)
                audio.export(file_str, format="mp3")
                self.window.ui.nodes['audio.transcribe.status'].setText("Converted: {}".format(os.path.basename(file_str)))
                path = file_str
            except Exception as e:
                self.window.core.debug.log(e)
                self.window.ui.nodes['audio.transcribe.status'].setText(
                    "Aborted. Can't convert video to mp3! FFMPEG not installed?\n"
                    "Please install \"ffmpeg\" or disable the option \"Always convert video to mp3\" to transcribe from video file.")
                self.window.ui.dialogs.alert(e)
                path = None
        return path

    def toggle_auto_convert(self):
        """Toggle auto video conversion"""
        state = self.window.ui.nodes['audio.transcribe.convert_video'].isChecked()
        self.window.core.config.set('audio.transcribe.convert_video', state)
        self.window.core.config.save()

    def restore_auto_convert(self):
        """Restore auto video conversion"""
        if self.window.core.config.has('audio.transcribe.convert_video'):
            state = self.window.core.config.get('audio.transcribe.convert_video', True)
            self.window.ui.nodes['audio.transcribe.convert_video'].setChecked(state)

    def is_auto_convert(self) -> bool:
        """
        Check if auto video conversion is enabled

        :return: True if enabled
        """
        return self.window.ui.nodes['audio.transcribe.convert_video'].isChecked()

    def store(self, text: str):
        """
        Store transcription to file

        :param text: transcribed text
        """
        path = os.path.join(self.window.core.config.get_user_path(), "transcript.txt")
        with open(path, "w", encoding="utf-8") as f:
            f.write(text)

    def store_current(self):
        """Store current transcription to file"""
        text = self.window.ui.editor["audio.transcribe"].toPlainText()
        self.store(text)

    def restore(self):
        """Restore transcription from file"""
        data = ""
        path = os.path.join(self.window.core.config.get_user_path(), "transcript.txt")
        if os.path.exists(path):
            with open(path, "r", encoding="utf-8") as f:
                try:
                    data = f.read()
                except:
                    pass
                self.window.ui.editor["audio.transcribe"].setPlainText(data)

    def is_video(self, file: str) -> bool:
        """
        Check if file is a video.

        :param file: file path
        :return: True if file is a video
        """
        ext = file.split(".")[-1].lower()
        return ext in self.video_extensions

    def show_hide(self, show: bool = True):
        """
        Show/hide transcribe window

        :param show: show/hide
        """
        if show:
            self.open()
        else:
            self.close()

    def on_close(self):
        """On transcribe window close"""
        self.opened = False
        self.update()

    def on_exit(self):
        """On exit"""
        self.store_current()

    def clear(self, force: bool = False):
        """
        Clear transcribe data

        :param force: force clear
        """
        if not force:
            self.window.ui.dialogs.confirm(
                type='audio.transcribe.clear',
                id=0,
                msg=trans("audio.transcribe.clear.confirm"),
            )
            return
        id = 'audio.transcribe'
        self.window.ui.editor[id].clear()
        self.window.ui.nodes['audio.transcribe.status'].setText("")
        self.store("")

    def setup_menu(self) -> dict:
        """
        Setup main menu

        :return dict with menu actions
        """
        actions = {}
        actions["audio.transcribe"] = QAction(
            QIcon(":/icons/hearing.svg"),
            trans("menu.tools.audio.transcribe"),
            self.window,
            checkable=False,
        )
        actions["audio.transcribe"].triggered.connect(
            lambda: self.toggle()
        )
        return actions

    def setup_dialogs(self):
        """Setup dialogs (static)"""
        self.dialog = AudioTranscribe(self.window)
        self.dialog.setup()

    def get_lang_mappings(self) -> dict:
        """
        Get language mappings

        :return: dict with language mappings
        """
        return {
            'menu.text': {
                'tools.audio.transcribe': 'menu.tools.audio.transcribe',
            }
        }


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.04.20 06:00:00                  #
# ================================================== #


class Finder:
    def __init__(self, window=None):
        """
        Finder controller

        :param window: Window instance
        """
        self.window = window
        self.parent = None
        self.locked = False
        self.opened = False

    def clear_search(self):
        """Clear search state"""
        if self.parent is not None:
            self.parent.clear_search()
        self.update_counter()

    def focus_input(self, text: str):
        """
        On focus in search string input

        :param text: search string
        """
        if text is not None and len(text) > 0:
            if self.parent is not None:
                self.parent.find(text)
        self.update_counter()

    def focus_in(self, parent):
        """
        On focus in parent

        :param parent: parent
        """
        self.set(parent)
        self.update_counter()

    def focus_out(self, parent):
        """
        On focus out parent

        :param parent: parent
        """
        pass

    def search_text_changed(self, search_string: str):
        """
        On search input text changed

        :param search_string: search string
        """
        if self.parent is not None:
            self.parent.find(search_string)
        self.update_counter()

    def prev(self):
        """Find previous match"""
        if self.parent is not None:
            self.parent.find_prev()
        self.update_counter()

    def next(self):
        """Find next match"""
        if self.parent is not None:
            self.parent.find_next()
        self.update_counter()

    def get_search_string(self):
        """
        Get current search string from input

        :return: search string
        """
        return self.window.ui.nodes['dialog.find.input'].text()

    def set(self, parent):
        """
        Set current parent for the finder

        :param parent: parent
        """
        self.parent = parent

    def unset(self, parent):
        """
        Unset parent from the finder

        :param parent: parent
        """
        if self.parent == parent:
            self.parent = None

    def open(self, parent):
        """
        Open finder dialog

        :param parent: parent finder
        """
        self.parent = parent
        if self.opened:
            self.window.ui.nodes['dialog.find.input'].setFocus()
            return
        if self.parent is not None:
            self.parent.prepare(clear=False)
        self.window.ui.dialog['find'].show()
        self.opened = True
        self.window.ui.nodes['dialog.find.input'].setFocus()
        current = self.get_search_string()
        if current is not None and len(current) > 0:
            if self.parent is not None:
                self.parent.find(current)

    def close(self, reset: bool = True):
        """
        Close finder dialog

        :param reset: Reset highlights
        """
        if reset and self.parent is not None:
            self.parent.reset()
        self.window.ui.dialog['find'].hide()
        self.opened = False

    def clear(self, restore: bool = False, to_end: bool = True):
        """
        Clear current parent highlights

        :param restore: Restore original content
        :param to_end: Move cursor to the end
        """
        if self.parent is not None:
            self.parent.clear(restore, to_end)
        self.update_counter()

    def clear_input(self):
        """Clear input text"""
        self.window.ui.nodes['dialog.find.input'].clear()
        if self.parent is not None:
            self.parent.find("")

    def update_counter(self):
        """Update counter for current parent"""
        if not self.opened:
            return
        self.window.ui.nodes['dialog.find.counter'].setText("0/0")
        if self.parent is None:
            return
        current = 0
        total_found = 0
        if isinstance(self.parent.matches, list):
            current = self.parent.current_match_index + 1 if self.parent.current_match_index >= 0 else 0
            total_found = len(self.parent.matches)
        elif isinstance(self.parent.matches, int):
            current = self.parent.current_match_index if self.parent.current_match_index >= 0 else 0
            total_found = self.parent.matches
        self.window.ui.nodes['dialog.find.counter'].setText("{}/{}".format(current, total_found))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.04.19 01:00:00                  #
# ================================================== #

import os

from PySide6.QtGui import QTextCursor, QAction, QIcon

from pygpt_net.tools.base import BaseTool
from pygpt_net.tools.code_interpreter.ui.dialogs import Interpreter
from pygpt_net.core.dispatcher import Event
from pygpt_net.item.ctx import CtxItem
from pygpt_net.utils import trans


class CodeInterpreter(BaseTool):
    def __init__(self, *args, **kwargs):
        """
        Python code interpreter

        :param window: Window instance
        """
        super(CodeInterpreter, self).__init__(*args, **kwargs)
        self.id = "interpreter"
        self.opened = False
        self.is_edit = False
        self.auto_clear = True
        self.dialog = None

        # interpreter data files in /data directory
        self.file_current = ".interpreter.current.py"
        self.file_input = ".interpreter.input.py"
        self.file_output = ".interpreter.output.py"

    def setup(self):
        """Setup"""
        self.load_history()
        self.load_output()
        self.update()

        # restore
        if self.window.core.config.has("interpreter.input"):
            self.window.ui.nodes['interpreter.input'].setPlainText(self.window.core.config.get("interpreter.input"))
        if self.window.core.config.has("interpreter.execute_all"):
            self.window.ui.nodes['interpreter.all'].setChecked(self.window.core.config.get("interpreter.execute_all"))
        if self.window.core.config.has("interpreter.auto_clear"):
            self.window.ui.nodes['interpreter.auto_clear'].setChecked(self.window.core.config.get("interpreter.auto_clear"))

    def on_reload(self):
        """On app profile reload"""
        self.setup()

    def update(self):
        """Update menu"""
        self.update_menu()

    def update_menu(self):
        """Update menu"""
        """
        if self.opened:
            self.window.ui.menu['tools.interpreter'].setChecked(True)
        else:
            self.window.ui.menu['tools.interpreter'].setChecked(False)
        """

    def append_output(self, output: str, type="stdout", **kwargs):
        """
        Append output to the interpreter window

        :param output: Output data
        :param type: Output type
        :param kwargs: Additional parameters
        """
        area = self.window.interpreter
        if type == "stdin":
            data = ">> "  + str(output)
        else:
            data = str(output)
        cur = area.textCursor()
        cur.movePosition(QTextCursor.End)
        s = data + "\n"
        while s:
            head, sep, s = s.partition("\n")
            cur.insertText(head)
            if sep:  # New line if LF
                cur.insertText("\n")
        area.setTextCursor(cur)
        self.save_output()
        self.load_history()

    def load_history(self):
        """Load history data from file"""
        data = self.get_history()
        self.window.ui.nodes['interpreter.code'].setPlainText(data)
        self.cursor_to_end()

    def get_output(self) -> str:
        """
        Load and get output from file

        :return: Output data
        """
        data = ""
        path = os.path.join(self.window.core.config.get_user_dir("data"), self.file_output)
        if os.path.exists(path):
            with open(path, "r", encoding="utf-8") as f:
                try:
                    data = f.read()
                except:
                    pass
        return data

    def load_output(self):
        """Load output data from file"""
        data = self.get_output()
        self.window.interpreter.setPlainText(data)
        self.window.ui.nodes['interpreter.input'].setFocus()

    def save_output(self):
        """Save output data to file"""
        path = os.path.join(self.window.core.config.get_user_dir("data"), self.file_output)
        data = self.window.interpreter.toPlainText()
        with open(path, "w", encoding="utf-8") as f:
            f.write(data)

    def get_history(self) -> str:
        """
        Load and get history from file

        :return: History data
        """
        data = ""
        path = os.path.join(self.window.core.config.get_user_dir("data"), self.file_input)
        if os.path.exists(path):
            with open(path, "r", encoding="utf-8") as f:
                try:
                    data = f.read()
                except:
                    pass
        return data

    def save_history(self, input: str):
        """
        Save input data

        :param input: Input data
        """
        path = os.path.join(self.window.core.config.get_user_dir("data"), self.file_input)
        with open(path , "w", encoding="utf-8") as f:
            f.write(input)

    def clear_history(self):
        """Clear input"""
        path = os.path.join(self.window.core.config.get_user_dir("data"), self.file_input)
        if os.path.exists(path):
            os.remove(path)
        self.window.ui.nodes['interpreter.code'].clear()

    def clear_output(self):
        """Clear output"""
        path = os.path.join(self.window.core.config.get_user_dir("data"), self.file_output)
        if os.path.exists(path):
            os.remove(path)
        self.window.interpreter.clear()

    def clear(self, force: bool = False):
        """
        Clear current window

        :param force: Force clear
        """
        if not force:
            self.window.ui.dialogs.confirm(
                type='interpreter.clear',
                id=0,
                msg=trans("interpreter.clear.confirm"),
            )
            return
        self.clear_output()

    def clear_all(self):
        """Clear input and output"""
        self.clear_output()
        self.clear_history()

    def send_input(self):
        """Send input to interpreter"""
        self.store_history()
        if self.auto_clear:
            self.clear_output()

        input = str(self.window.ui.nodes['interpreter.input'].toPlainText())
        if self.is_all():
            cmd = "code_execute_all"
        else:
            if input == "":
                return
            cmd = "code_execute"

        commands = [
            {
                "cmd": cmd,
                "params": {
                    "code": input,
                    "path": self.file_current,
                },
                "silent": True,
            }
        ]
        event = Event(Event.CMD_EXECUTE, {
            'commands': commands,
        })
        event.ctx = CtxItem()  # tmp
        self.window.controller.command.dispatch_only(event)
        self.window.ui.nodes['interpreter.input'].clear()
        self.window.ui.nodes['interpreter.input'].setFocus()

    def update_input(self):
        """Update input data"""
        data = self.window.ui.nodes['interpreter.input'].toPlainText()
        self.window.core.config.set("interpreter.input", data)

    def append_to_input(self, data: str):
        """
        Append data to input

        :param data: Data
        """
        current = self.window.ui.nodes['interpreter.input'].toPlainText()
        if current:
            current += "\n"
        current += data
        self.window.ui.nodes['interpreter.input'].setPlainText(current)

    def append_to_edit(self, data: str):
        """
        Append data to editor

        :param data: Data
        """
        prev = self.get_history()
        if prev:
            prev += "\n"
        prev += data
        self.save_history(prev)
        self.load_history()

    def store_history(self):
        """Save history data to file"""
        data = self.window.ui.nodes['interpreter.code'].toPlainText()
        self.save_history(data)

    def open(self):
        """Open interpreter dialog"""
        self.opened = True
        self.load_history()
        self.load_output()
        self.window.ui.dialogs.open('interpreter', width=800, height=600)
        self.window.ui.nodes['interpreter.input'].setFocus()
        self.cursor_to_end()
        self.update()

    def close(self):
        """Close interpreter dialog"""
        self.opened = False
        self.window.ui.dialogs.close('interpreter')
        self.update()

    def toggle(self):
        """Toggle interpreter dialog open/close"""
        if self.opened:
            self.close()
        else:
            self.open()

    def show_hide(self, show: bool = True):
        """
        Show/hide interpreter window

        :param show: show/hide
        """
        if show:
            self.open()
        else:
            self.close()

    def toggle_icon(self, state: bool):
        """
        Toggle interpreter icon

        :param state: State
        """
        self.window.ui.nodes['icon.interpreter'].setVisible(state)

    def toggle_auto_clear(self):
        """Toggle auto clear"""
        self.auto_clear = self.window.ui.nodes['interpreter.auto_clear'].isChecked()
        self.window.core.config.set("interpreter.auto_clear", self.auto_clear)

    def toggle_all(self):
        """Toggle execute all"""
        self.window.core.config.set("interpreter.execute_all", self.window.ui.nodes['interpreter.all'].isChecked())

    def get_current_output(self) -> str:
        """
        Get current output from interpreter

        :return: Output data
        """
        return self.window.interpreter.toPlainText()

    def get_current_history(self) -> str:
        """
        Get code from history

        :return: Edit code
        """
        return self.window.ui.nodes['interpreter.code'].toPlainText()

    def is_all(self) -> bool:
        """
        Check if execute all is enabled

        :return: True if execute all is enabled
        """
        return self.window.ui.nodes['interpreter.all'].isChecked()

    def cursor_to_end(self):
        """Move cursor to end"""
        cur = self.window.ui.nodes['interpreter.code'].textCursor()
        cur.movePosition(QTextCursor.End)
        self.window.ui.nodes['interpreter.code'].setTextCursor(cur)

    def setup_menu(self) -> dict:
        """
        Setup main menu

        :return dict with menu actions
        """
        actions = {}
        actions["interpreter"] = QAction(
            QIcon(":/icons/code.svg"),
            trans("menu.tools.interpreter"),
            self.window,
            checkable=False,
        )
        actions["interpreter"].triggered.connect(
            lambda: self.toggle()
        )
        return actions

    def setup_dialogs(self):
        """Setup dialogs (static)"""
        self.dialog = Interpreter(self.window)
        self.dialog.setup()

    def setup_theme(self):
        """Setup theme"""
        size = self.window.core.config.get('font_size')
        self.window.interpreter.value = size

    def get_lang_mappings(self) -> dict:
        """
        Get language mappings

        :return: dict with language mappings
        """
        return {
            'menu.text': {
                'tools.interpreter': 'menu.tools.interpreter',
            }
        }


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.05.02 19:00:00                  #
# ================================================== #

import pyaudio
import wave
import os

from PySide6.QtCore import QTimer, Slot
from PySide6.QtWidgets import QApplication

from pygpt_net.item.ctx import CtxItem
from pygpt_net.plugin.audio_input.worker import ControlWorker
from pygpt_net.core.dispatcher import BaseEvent
from pygpt_net.core.access.events import ControlEvent, AppEvent
from pygpt_net.utils import trans


class Voice:

    TIMEOUT_SECONDS = 120  # 2 minutes, max recording time before timeout

    def __init__(self, window=None):
        """
        Accessibility voice controller

        :param window: Window instance
        """
        self.window = window
        self.is_recording = False
        self.frames = []
        self.p = None
        self.stream = None
        self.timer = None
        self.input_file = "voice_control.wav"
        self.thread_started = False
        self.audio_disabled_events = [
            AppEvent.INPUT_CALL,
            AppEvent.VOICE_CONTROL_TOGGLE,
        ]
        self.confirm_events = [
            ControlEvent.NOTEPAD_CLEAR,
            ControlEvent.CALENDAR_CLEAR,
        ]

    def setup(self):
        """Setup voice control"""
        self.update()

    def play(self, event: AppEvent):
        """
        Play audio event

        :param event: AppEvent
        """
        mic_events = [
            AppEvent.VOICE_CONTROL_STARTED,
            AppEvent.VOICE_CONTROL_STOPPED,
            AppEvent.VOICE_CONTROL_SENT,
            AppEvent.INPUT_VOICE_LISTEN_STARTED,
            AppEvent.INPUT_VOICE_LISTEN_STOPPED,
        ]
        always_play = [
            AppEvent.VOICE_CONTROL_UNRECOGNIZED,
        ]
        if event.name in mic_events and self.window.core.config.get("access.microphone.notify"):
            if event.name in [AppEvent.VOICE_CONTROL_STARTED, AppEvent.INPUT_VOICE_LISTEN_STARTED]:
                if self.window.core.config.get("access.audio.notify.execute"):
                    self.window.controller.audio.play_sound("click_on.mp3")
            elif event.name in [
                AppEvent.VOICE_CONTROL_STOPPED,
                AppEvent.VOICE_CONTROL_SENT,
                AppEvent.INPUT_VOICE_LISTEN_STOPPED
            ]:
                if self.window.core.config.get("access.audio.notify.execute"):
                    self.window.controller.audio.play_sound("click_off.mp3")
            return

        if not self.window.core.config.get("access.audio.event.speech") and event.name not in always_play:
            return

        if event.name not in self.audio_disabled_events:
            trans_key = "event.audio." + event.name
            if event.name == AppEvent.CTX_SELECTED:
                self.window.controller.audio.read_text(
                    self.window.core.access.voice.get_selected_ctx()  # with info about ctx
                )
            elif event.name == AppEvent.CTX_END:
                if not self.window.controller.plugins.is_type_enabled("audio.output"):
                    self.window.controller.audio.read_text(trans(trans_key))  # only if audio output is disabled
            else:
                # handle rest of events
                self.window.controller.audio.read_text(trans(trans_key))
            self.window.core.debug.info("AUDIO EVENT PLAY: " + event.name)

    def update(self):
        """Update voice control"""
        if self.window.core.config.get("access.voice_control"):
            self.window.ui.nodes['voice.control.btn'].setVisible(True)
        else:
            self.window.ui.nodes['voice.control.btn'].setVisible(False)

    def toggle_recording(self):
        """Toggle recording"""
        if self.is_recording:
            self.stop_recording()
        else:
            self.start_recording()

    def switch_btn_stop(self):
        """Switch button to stop"""
        self.window.ui.nodes['voice.control.btn'].btn_toggle.setText(trans('audio.speak.btn.stop'))
        self.window.ui.nodes['voice.control.btn'].btn_toggle.setToolTip(
            trans('audio.speak.btn.stop.tooltip'))

    def switch_btn_start(self):
        """Switch button to start"""
        self.window.ui.nodes['voice.control.btn'].btn_toggle.setText(trans('audio.control.btn'))
        self.window.ui.nodes['voice.control.btn'].btn_toggle.setToolTip(trans('audio.speak.btn.tooltip'))

    def stop_timeout(self):
        """Stop timeout"""
        self.stop_recording(timeout=True)

    def start_recording(self):
        """Start recording"""
        self.frames = []  # clear audio frames

        def callback(in_data, frame_count, time_info, status):
            self.frames.append(in_data)
            if self.is_recording:
                return (in_data, pyaudio.paContinue)
            else:
                return (in_data, pyaudio.paComplete)

        try:
            self.is_recording = True
            self.switch_btn_stop()

            # start timeout timer to prevent infinite recording
            if self.timer is None:
                self.timer = QTimer()
                self.timer.timeout.connect(self.stop_timeout)
                self.timer.start(self.TIMEOUT_SECONDS * 1000)

            self.p = pyaudio.PyAudio()
            self.stream = self.p.open(format=pyaudio.paInt16,
                                      channels=1,
                                      rate=44100,
                                      input=True,
                                      frames_per_buffer=1024,
                                      stream_callback=callback)

            self.window.ui.status(trans('audio.speak.now'))
            self.window.core.dispatcher.dispatch(AppEvent(AppEvent.VOICE_CONTROL_STARTED))  # app event
            self.stream.start_stream()
        except Exception as e:
            self.is_recording = False
            self.window.core.debug.log(e)
            self.window.ui.dialogs.alert(e)
            if self.window.core.platforms.is_snap():
                self.window.ui.dialogs.open(
                    'snap_audio_input',
                    width=400,
                    height=200
                )
            self.switch_btn_start()  # switch button to start

    def stop_recording(self, timeout: bool = False):
        """
        Stop recording

        :param timeout: True if stopped due to timeout
        """
        self.is_recording = False
        if self.timer:
            self.timer.stop()
            self.timer = None
        self.switch_btn_start()  # switch button to start
        path = os.path.join(self.window.core.config.path, self.input_file)

        if self.stream is not None:
            self.stream.stop_stream()
            self.stream.close()
            self.p.terminate()

            # abort if timeout
            if timeout:
                self.window.core.dispatcher.dispatch(AppEvent(AppEvent.VOICE_CONTROL_STOPPED))  # app event
                self.window.ui.status("Aborted.".format(self.TIMEOUT_SECONDS))
                return

            if self.frames:
                wf = wave.open(path, 'wb')
                wf.setnchannels(1)
                wf.setsampwidth(self.p.get_sample_size(pyaudio.paInt16))
                wf.setframerate(44100)
                wf.writeframes(b''.join(self.frames))
                wf.close()
                self.window.core.dispatcher.dispatch(AppEvent(AppEvent.VOICE_CONTROL_SENT))  # app event
                self.handle_thread(True)  # handle transcription in simple mode
        else:
            self.window.core.dispatcher.dispatch(AppEvent(AppEvent.VOICE_CONTROL_STOPPED))  # app event

    def handle_thread(self, force: bool = False):
        """
        Handle listener thread

        :param force: force start
        """
        if self.thread_started and not force:
            return

        try:
            # worker
            worker = ControlWorker()
            worker.window = self.window
            worker.path = os.path.join(
                self.window.core.config.path,
                self.input_file,
            )
            # signals
            worker.signals.transcribed.connect(self.handle_transcribed)
            worker.signals.finished.connect(self.handle_input)
            worker.signals.destroyed.connect(self.handle_destroy)
            worker.signals.started.connect(self.handle_started)
            worker.signals.stopped.connect(self.handle_stop)
            worker.signals.status.connect(self.handle_status)
            worker.signals.error.connect(self.handle_error)

            # start
            self.window.threadpool.start(worker)

        except Exception as e:
            self.window.core.debug.log(e)

    @Slot(str, str)
    def handle_transcribed(self, path: str, text: str):
        """
        Handle transcribed text

        :param path: audio file path
        :param text: transcribed text
        """
        self.window.tools.get("transcriber").on_transcribe(path, text)

    @Slot(object, object)
    def handle_input(self, text: str, ctx: CtxItem = None):
        """
        Insert text to input and send

        :param text: text
        :param ctx: CtxItem
        """
        if text is None or text.strip() == '':
            return
        self.window.core.debug.info("VOICE CONTROL INPUT: " + text)
        commands = self.window.core.access.voice.recognize_commands(text)
        if len(commands) > 0:
            for command in commands:
                cmd = command["cmd"]
                params = command.get("params", "")
                self.window.ui.status("Voice command: " + cmd)
                event = ControlEvent(cmd)
                event.data = {
                    "params": params,
                }
                self.window.core.dispatcher.dispatch(event)
                self.window.core.debug.info("VOICE CONTROL COMMAND: " + cmd, params)
                trans_key = "event.control." + cmd

                if cmd in self.confirm_events:
                    self.window.ui.status(trans("event.audio.confirm"))
                else:
                    self.window.ui.status("Voice command: " + trans(trans_key))
                    QApplication.processEvents()

            # play OK sound
            if self.window.core.config.get("access.audio.notify.execute"):
                self.window.controller.audio.play_sound("ok.mp3")

        else:
            self.window.ui.status("Voice command not recognized.")
            QApplication.processEvents()
            self.window.core.dispatcher.dispatch(AppEvent(AppEvent.VOICE_CONTROL_UNRECOGNIZED))

    @Slot(object)
    def handle_error(self, data: str):
        """
        Handle thread error signal

        :param data: message
        """
        self.window.ui.status(str(data))

    @Slot(object)
    def handle_status(self, data: str):
        """
        Handle thread status msg signal

        :param data: message
        """
        self.window.ui.status(str(data))

    @Slot()
    def handle_destroy(self):
        """Handle listener destroyed"""
        self.thread_started = False

    @Slot()
    def handle_started(self):
        """Handle listening started"""
        self.thread_started = True

    @Slot()
    def handle_stop(self):
        """Handle stop listening"""
        self.thread_started = False

    def read_tab_name(self):
        """Read audio tab name"""
        text = self.window.core.access.voice.get_selected_tab()
        if self.window.core.config.get("access.audio.event.speech"):
            self.window.controller.audio.read_text(text)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.04.30 16:00:00                  #
# ================================================== #

import json

from pygpt_net.item.assistant import AssistantItem
from pygpt_net.item.ctx import CtxItem

from .worker.assistants import AssistantsWorker, EventHandler
from .worker.importer import Importer


class Assistants:
    def __init__(self, window=None):
        """
        GPT Assistants API Wrapper

        :param window: Window instance
        """
        self.window = window
        self.worker = AssistantsWorker(window)
        self.importer = Importer(window)

    def get_client(self):
        """
        Get OpenAI client

        :return: OpenAI client
        """
        return self.window.core.gpt.get_client()

    def log(self, msg: str, callback: callable = None):
        """
        Log message

        :param msg: message to log
        :param callback: callback log function
        """
        if callback is not None:
            callback(msg)
        else:
            print(msg)

    def create(self, assistant: AssistantItem) -> AssistantItem:
        """
        Create assistant

        :param assistant: assistant object
        :return: assistant object
        """
        client = self.get_client()
        tools = self.get_tools(assistant)
        tool_resources = self.get_tool_resources(assistant)
        result = client.beta.assistants.create(
            instructions=assistant.instructions,
            description=assistant.description,
            name=assistant.name,
            tools=tools,
            model=assistant.model,
            tool_resources=tool_resources,
        )
        if result is not None:
            assistant.id = result.id
            return assistant

    def update(self, assistant: AssistantItem) -> AssistantItem:
        """
        Update assistant

        :param assistant: assistant object
        :return: assistant object
        """
        client = self.get_client()
        tools = self.get_tools(assistant)
        tool_resources = self.get_tool_resources(assistant)
        result = client.beta.assistants.update(
            assistant.id,
            instructions=assistant.instructions,
            description=assistant.description,
            name=assistant.name,
            tools=tools,
            model=assistant.model,
            tool_resources=tool_resources,
        )
        if result is not None:
            assistant.id = result.id
            return assistant

    def delete(self, id: str) -> str:
        """
        Delete assistant by ID

        :param id: assistant ID
        :return: assistant ID
        """
        client = self.get_client()
        response = client.beta.assistants.delete(id)
        if response is not None:
            return response.id

    def thread_create(self) -> str:
        """
        Create thread and return thread ID

        :return: thread ID
        """
        client = self.get_client()
        thread = client.beta.threads.create()
        if thread is not None:
            return thread.id

    def thread_delete(self, id: str) -> str:
        """
        Delete thread by thread ID

        :param id: thread ID
        :return: thread ID
        """
        client = self.get_client()
        response = client.beta.threads.delete(id)
        if response is not None:
            return response.id

    def msg_list(self, thread_id: str) -> list:
        """
        Get messages from thread

        :param thread_id: thread ID
        :return: messages data list
        """
        client = self.get_client()
        thread_messages = client.beta.threads.messages.list(thread_id)
        return thread_messages.data

    def msg_send(self, id: str, text: str, file_ids: list = None):
        """
        Send message to thread

        :param id: thread ID
        :param text: message text
        :param file_ids: uploaded files IDs
        :return: message
        """
        client = self.get_client()
        additional_args = {}
        attachments = []
        for file_id in file_ids:  # append file ids from attachments
            attachments.append({
                "file_id": file_id,
                "tools": [
                    {
                        "type": "code_interpreter"
                    }, {
                        "type": "file_search"
                    },
                ]
            })
        if attachments:
            additional_args['attachments'] = attachments

        message = client.beta.threads.messages.create(
            id,
            role="user",
            content=text,
            **additional_args
        )
        if message is not None:
            return message

    def get_tools(self, assistant: AssistantItem) -> list:
        """
        Get assistant tools

        :param assistant: assistant item
        :return: tools list
        """
        tools = []
        if assistant.has_tool("code_interpreter"):
            tools.append({"type": "code_interpreter"})
        if assistant.has_tool("file_search"):
            tools.append({"type": "file_search"})

        # functions
        if assistant.has_functions():
            functions = assistant.get_functions()
            for function in functions:
                if str(function['name']).strip() == '' or function['name'] is None:
                    continue
                params = json.loads(function['params'])  # unpack JSON from string
                tools.append(
                    {
                        "type": "function",
                        "function": {
                            "name": function['name'],
                            "parameters": params,
                            "description": function['desc'],
                        }
                    }
                )
        return tools

    def get_tool_resources(self, assistant: AssistantItem) -> dict:
        """
        Get assistant tool resources

        :param assistant: assistant item
        :return: tool resources dict
        """
        resources = {}
        if assistant.has_tool("file_search"):
            if assistant.vector_store is not None and assistant.vector_store != "":
                resources["file_search"] = {
                    "vector_store_ids": [assistant.vector_store]
                }
        return resources

    def run_create(
            self,
            thread_id: str,
            assistant_id: str,
            model: str = None,
            instructions: str = None
    ):
        """
        Create assistant run

        :param thread_id: thread ID
        :param assistant_id: assistant ID
        :param model: model
        :param instructions: instructions
        :return: Run
        """
        client = self.get_client()
        args = {}
        if instructions is not None and instructions != "":
            args['instructions'] = instructions
        if model is not None:
            args['model'] = model

        run = client.beta.threads.runs.create(
            thread_id=thread_id,
            assistant_id=assistant_id,
            **args
        )
        if run is not None:
            return run

    def run_create_stream(
            self,
            signals,
            ctx: CtxItem,
            thread_id: str,
            assistant_id: str,
            model: str = None,
            instructions: str = None
    ):
        """
        Create assistant run (stream)

        :param signals: worker signals (callbacks)
        :param ctx: context item
        :param thread_id: thread ID
        :param assistant_id: assistant ID
        :param model: model
        :param instructions: instructions
        :return: Run (final)
        """
        client = self.get_client()
        additional_args = {}
        if instructions is not None and instructions != "":
            additional_args['instructions'] = instructions
        if model is not None:
            additional_args['model'] = model

        with client.beta.threads.runs.stream(
                thread_id=thread_id,
                assistant_id=assistant_id,
                event_handler=EventHandler(signals, ctx),
                **additional_args
        ) as stream:
            return stream.get_final_run()

    def run_status(
            self,
            ctx: CtxItem
    ) -> str:
        """
        Get assistant run status

        :param ctx: context item
        :return: run status
        """
        client = self.get_client()
        run = client.beta.threads.runs.retrieve(
            thread_id=ctx.thread,
            run_id=ctx.run_id,
        )
        if run is not None:
            if run.usage is not None:
                ctx.input_tokens = run.usage["prompt_tokens"]
                ctx.output_tokens = run.usage["completion_tokens"]
                ctx.total_tokens = run.usage["total_tokens"]
            return run.status

    def run_get(
            self,
            ctx: CtxItem
    ) -> str:
        """
        Get assistant run status

        :param ctx: context item
        :return: run
        """
        client = self.get_client()
        run = client.beta.threads.runs.retrieve(
            thread_id=ctx.thread,
            run_id=ctx.run_id,
        )
        return run

    def run_stop(
            self,
            ctx: CtxItem
    ) -> str:
        """
        Stop assistant run

        :param ctx: context item
        :return: run status
        """
        client = self.get_client()
        run = client.beta.threads.runs.cancel(
            thread_id=ctx.thread,
            run_id=ctx.run_id,
        )
        if run is not None:
            if run.usage is not None:
                ctx.input_tokens = run.usage["prompt_tokens"]
                ctx.output_tokens = run.usage["completion_tokens"]
                ctx.total_tokens = run.usage["total_tokens"]
            return run.status

    def run_submit_tool(
            self,
            ctx: CtxItem,
            outputs: list,
    ):
        """
        Submit assistant run tool outputs

        :param ctx: context item
        :param outputs: outputs
        :return: run
        """
        client = self.get_client()
        return client.beta.threads.runs.submit_tool_outputs(
            thread_id=ctx.thread,
            run_id=ctx.run_id,
            tool_outputs=outputs,
        )

    def import_all(
            self,
            items: dict,
            order: str = "asc",
            limit: int = 100,
            after: str = None,
            callback: callable = None
    ) -> dict:
        """
        Import assistants from API (all, paginated)

        :param items: items
        :param order: order
        :param limit: limit
        :param after: next page after ID
        :param callback: callback log function
        :return: items dict
        """
        client = self.get_client()
        args = {
            "order": order,
            "limit": limit,
        }
        if after is not None:
            args['after'] = after
        assistants = client.beta.assistants.list(**args)
        if assistants is not None:
            for remote in assistants.data:
                id = remote.id
                if id not in items:
                    items[id] = AssistantItem()
                items[id].id = id
                items[id].name = remote.name
                items[id].description = remote.description
                items[id].instructions = remote.instructions
                items[id].model = remote.model
                items[id].meta = remote.metadata

                # reset tools
                items[id].clear_tools()
                items[id].vector_store = ""

                # tool resources
                if (hasattr(remote.tool_resources, 'file_search')
                        and remote.tool_resources.file_search is not None):
                    # get vector store ID
                    if remote.tool_resources.file_search.vector_store_ids:
                        items[id].vector_store = remote.tool_resources.file_search.vector_store_ids[0]

                # append tools
                for tool in remote.tools:
                    if tool.type == "function":
                        # pack params to JSON string
                        params = ''
                        try:
                            params = json.dumps(
                                tool.function.parameters
                            )
                        except:
                            pass
                        items[id].add_function(
                            tool.function.name,
                            params,
                            tool.function.description
                        )
                    else:
                        if tool.type in items[id].tools:
                            items[id].tools[tool.type] = True

                self.log("Imported assistant: " + remote.id, callback)

            # next page
            if assistants.has_more:
                return self.import_all(
                    items,
                    order,
                    limit,
                    assistants.last_id,
                    callback,
                )
        return items


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.05.02 19:00:00                  #
# ================================================== #

from PySide6.QtCore import Qt
from PySide6.QtGui import QStandardItemModel
from PySide6.QtWidgets import QPushButton, QHBoxLayout, QLabel, QVBoxLayout, QScrollArea, QWidget, QSizePolicy, \
    QTabWidget, QSplitter

from pygpt_net.ui.base.config_dialog import BaseConfigDialog
from pygpt_net.ui.widget.dialog.settings import SettingsDialog
from pygpt_net.ui.widget.element.group import CollapsedGroup
from pygpt_net.ui.widget.lists.settings import SettingsSectionList
from pygpt_net.utils import trans


class Settings(BaseConfigDialog):
    def __init__(self, window=None, *args, **kwargs):
        super(Settings, self).__init__(window, *args, **kwargs)
        """
        Settings (main) dialog

        :param window: Window instance
        """
        self.window = window
        self.id = "config"

    def setup(self, idx=None):
        """Setup settings dialog"""
        id = "settings"
        path = self.window.core.config.path
        sections = self.window.core.settings.get_sections()

        # buttons
        self.window.ui.nodes['settings.btn.defaults.user'] = QPushButton(trans("dialog.settings.btn.defaults.user"))
        self.window.ui.nodes['settings.btn.defaults.app'] = QPushButton(trans("dialog.settings.btn.defaults.app"))
        self.window.ui.nodes['settings.btn.save'] = QPushButton(trans("dialog.settings.btn.save"))
        self.window.ui.nodes['settings.btn.defaults.user'].clicked.connect(
            lambda: self.window.controller.settings.editor.load_defaults_user())
        self.window.ui.nodes['settings.btn.defaults.app'].clicked.connect(
            lambda: self.window.controller.settings.editor.load_defaults_app())
        self.window.ui.nodes['settings.btn.save'].clicked.connect(
            lambda: self.window.controller.settings.editor.save(id))

        # set enter key to save button
        self.window.ui.nodes['settings.btn.defaults.user'].setAutoDefault(False)
        self.window.ui.nodes['settings.btn.defaults.app'].setAutoDefault(False)
        self.window.ui.nodes['settings.btn.save'].setAutoDefault(True)

        # bottom buttons layout
        bottom = QHBoxLayout()
        bottom.addWidget(self.window.ui.nodes['settings.btn.defaults.user'])
        bottom.addWidget(self.window.ui.nodes['settings.btn.defaults.app'])
        bottom.addWidget(self.window.ui.nodes['settings.btn.save'])

        self.window.ui.paths[id] = QLabel(str(path))
        self.window.ui.paths[id].setStyleSheet("font-weight: bold;")

        # settings section tabs
        self.window.ui.tabs['settings.section'] = QTabWidget()

        # build settings tabs
        for section_id in sections:
            content_tabs = {}
            scroll_tabs = {}
            advanced_keys = {}  # advanced options keys

            # get settings options for section
            fields = self.window.controller.settings.editor.get_options(section_id)

            # extract tab ids, general is default
            tab_ids = self.extract_option_tabs(fields)
            extra_tabs = self.append_tabs(section_id)
            if extra_tabs:
                tab_ids += extra_tabs
            for tab_id in tab_ids:
                content_tabs[tab_id] = QVBoxLayout()
                scroll_tabs[tab_id] = QScrollArea()
                scroll_tabs[tab_id].setWidgetResizable(True)

            # prepare advanced options keys
            for key in fields:
                if 'advanced' in fields[key] and fields[key]['advanced']:
                    tab_id = "general"
                    if 'tab' in fields[key]:
                        tab = fields[key]['tab']
                        if tab is not None and tab != "":
                            tab_id = tab
                    if tab_id not in advanced_keys:
                        advanced_keys[tab_id] = []
                    advanced_keys[tab_id].append(key)

            # build settings widgets
            widgets = self.build_widgets(self.id, fields)

            # apply settings widgets
            for key in widgets:
                self.window.ui.config[self.id][key] = widgets[key]

            # apply widgets to layouts
            options = {}
            for key in widgets:
                if fields[key]["type"] == 'text' or fields[key]["type"] == 'int' or fields[key]["type"] == 'float':
                    options[key] = self.add_option(widgets[key], fields[key])
                elif fields[key]["type"] == 'textarea':
                    options[key] = self.add_row_option(widgets[key], fields[key])
                elif fields[key]["type"] == 'bool':
                    options[key] = self.add_raw_option(widgets[key], fields[key])
                elif fields[key]['type'] == 'dict':
                    options[key] = self.add_row_option(widgets[key], fields[key])  # dict
                    # register dict to editor:
                    self.window.ui.dialogs.register_dictionary(
                        key,
                        parent="config",
                        option=fields[key],
                    )
                elif fields[key]['type'] == 'combo':
                    options[key] = self.add_option(widgets[key], fields[key])  # combobox

            self.window.ui.nodes['settings.api_key.label'].setMinimumHeight(60)

            # append widgets options layouts to scroll area
            for key in options:
                option = options[key]
                tab_id = "general"
                if 'tab' in fields[key]:
                    tab = fields[key]['tab']
                    if tab is not None and tab != "":
                        tab_id = tab

                # hide advanced options
                if tab_id in advanced_keys and key in advanced_keys[tab_id]:
                    continue

                content_tabs[tab_id].addLayout(option)  # add

                # append URLs
                if 'urls' in fields[key]:
                    content_tabs[tab_id].addWidget(self.add_urls(fields[key]['urls']))

                # check if not last option
                if key != list(options.keys())[-1] or tab_id in advanced_keys:
                    content_tabs[tab_id].addWidget(self.add_line())

            # append advanced options at the end
            if len(advanced_keys) > 0:
                groups = {}
                for key in options:
                    tab_id = "general"
                    if 'tab' in fields[key]:
                        tab = fields[key]['tab']
                        if tab is not None and tab != "":
                            tab_id = tab

                    if tab_id not in advanced_keys:
                        continue

                    # ignore non-advanced options
                    if key not in advanced_keys[tab_id]:
                        continue

                    group_id = 'settings.advanced.' + section_id + '.' + tab_id

                    if tab_id not in groups:
                        groups[tab_id] = CollapsedGroup(self.window, group_id, None, False, None)
                        groups[tab_id].box.setText(trans('settings.advanced.collapse'))

                    groups[tab_id].add_layout(options[key])  # add option to group

                    # add line if not last option
                    if key != advanced_keys[tab_id][-1]:
                        groups[tab_id].add_widget(self.add_line())

                # add advanced options group to scrolls
                for tab_id in groups:
                    group_id = 'settings.advanced.' + section_id + '.' + tab_id
                    content_tabs[tab_id].addWidget(groups[tab_id])
                    self.window.ui.groups[group_id] = groups[tab_id]

            # add extra features buttons
            self.append_extra(content_tabs, section_id, options, fields)

            # add stretch to the end of every option tab
            for tab_id in content_tabs:
                content_tabs[tab_id].addStretch()

            # tabs or no tabs
            if len(content_tabs) > 1:
                # tabs
                tab_widget = QTabWidget()

                # sort to make general tab first if exists
                if "general" in content_tabs:
                    content_tabs = {"general": content_tabs.pop("general")} | content_tabs

                for tab_id in content_tabs:
                    if tab_id == "general":
                        name_key = trans("settings.section.tab.general")
                    else:
                        name_key = trans("settings.section." + section_id + "." + tab_id)
                    tab_name = trans(name_key)
                    scroll_widget = QWidget()
                    scroll_widget.setLayout(content_tabs[tab_id])
                    scroll_tabs[tab_id].setWidget(scroll_widget)
                    tab_widget.addTab(scroll_tabs[tab_id], tab_name)

                area = QVBoxLayout()
                area.addWidget(self.add_line())
                area.addWidget(tab_widget)
            else:
                # one scroll only
                scroll_widget = QWidget()
                scroll_widget.setLayout(content_tabs["general"])
                scroll_tabs["general"].setWidget(scroll_widget)

                area = QVBoxLayout()
                area.addWidget(self.add_line())
                area.addWidget(scroll_tabs["general"])

            area_widget = QWidget()
            area_widget.setLayout(area)

            # append to tab
            self.window.ui.tabs['settings.section'].addTab(
                area_widget,
                trans(sections[section_id]['label'])
            )

        # -----------------------

        self.window.ui.tabs['settings.section'].currentChanged.connect(
            lambda: self.window.controller.settings.set_by_tab(
                self.window.ui.tabs['settings.section'].currentIndex())
        )

        data = {}
        for section_id in sections:
            section = sections[section_id]
            data[section_id] = section

        # section list
        data_id = 'settings.section.list'
        self.window.ui.nodes[data_id] = SettingsSectionList(self.window, data_id)
        self.window.ui.models[data_id] = self.create_model(self.window)
        self.window.ui.nodes[data_id].setModel(self.window.ui.models[data_id])

        # update section list
        self.update_list(data_id, data)

        # set max width to list
        self.window.ui.nodes[data_id].setMinimumWidth(200)

        # splitter
        self.window.ui.splitters['dialog.settings'] = QSplitter(Qt.Horizontal)
        self.window.ui.splitters['dialog.settings'].addWidget(self.window.ui.nodes[data_id])  # list
        self.window.ui.splitters['dialog.settings'].addWidget(self.window.ui.tabs['settings.section'])  # tabs
        self.window.ui.splitters['dialog.settings'].setStretchFactor(0, 2)
        self.window.ui.splitters['dialog.settings'].setStretchFactor(1, 5)
        self.window.ui.splitters['dialog.settings'].setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)

        main_layout = QHBoxLayout()
        main_layout.addWidget(self.window.ui.splitters['dialog.settings'])

        layout = QVBoxLayout()
        layout.addLayout(main_layout)  # list + plugins tabs
        layout.addLayout(bottom)  # bottom buttons (save, defaults)

        self.window.ui.dialog['config.' + id] = SettingsDialog(self.window, id)
        self.window.ui.dialog['config.' + id].setLayout(layout)
        self.window.ui.dialog['config.' + id].setWindowTitle(trans('dialog.settings'))

        # restore current opened tab if idx is set
        if idx is not None:
            try:
                self.window.ui.tabs['settings.section'].setCurrentIndex(idx)
                self.window.controller.settings.set_by_tab(idx)
            except Exception as e:
                print('Failed restore settings tab: {}'.format(idx))
        else:
            self.window.controller.settings.set_by_tab(0)

    def extract_option_tabs(self, options: dict) -> list:
        """
        Get keys for option tabs

        :param options: settings options
        :return: list with keys
        """
        keys = []
        is_default = False
        for key in options:
            option = options[key]
            if 'tab' in option:
                tab = option['tab']
                if tab == "" or tab is None:
                    is_default = True
                if tab not in keys:
                    keys.append(tab)
            else:
                is_default = True

        # add default general tab if not exists
        if len(keys) == 0 or (is_default and "general" not in keys):
            keys.append("general")
        return keys

    def append_extra(self, content: dict, section_id: str, widgets: dict, options: dict):
        """
        Append extra fields in real-time

        :param content: dict with content tabs
        :param section_id: section id
        :param widgets: options dict
        :param options: options dist
        """
        if section_id == 'llama-index':
            content["update"].addWidget(self.add_line())
            self.window.controller.idx.settings.append(content, widgets, options)

    def append_tabs(self, section_id: str) -> list:
        """
        Append extra tabs in real-time

        :param section_id: section id
        :return: list with tabs
        """
        if section_id == 'llama-index':
            return self.window.controller.idx.settings.append_tabs()
        return []

    def create_model(self, parent) -> QStandardItemModel:
        """
        Create list model

        :param parent: parent widget
        :return: QStandardItemModel
        """
        return QStandardItemModel(0, 1, parent)

    def update_list(self, id: str, data: dict):
        """
        Update list

        :param id: ID of the list
        :param data: Data to update
        """
        self.window.ui.models[id].removeRows(0, self.window.ui.models[id].rowCount())
        i = 0
        for n in data:
            self.window.ui.models[id].insertRow(i)
            name = trans(data[n]['label'])
            self.window.ui.models[id].setData(self.window.ui.models[id].index(i, 0), name)
            i += 1

    def refresh_list(self):
        """
        Refresh list

        :param id: ID of the list
        """
        sections = self.window.core.settings.get_sections()
        data = {}
        for section_id in sections:
            section = sections[section_id]
            data[section_id] = section

        data_id = 'settings.section.list'

        # update section list
        self.update_list(data_id, data)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.05.02 19:00:00                  #
# ================================================== #

from PySide6.QtWidgets import QHBoxLayout, QWidget, QComboBox

from pygpt_net.utils import trans


class NoScrollCombo(QComboBox):
    def __init__(self, parent=None):
        super(NoScrollCombo, self).__init__(parent)

    def wheelEvent(self, event):
        event.ignore()  # disable mouse wheel


class OptionCombo(QWidget):
    def __init__(self, window=None, parent_id: str = None, id: str = None, option: dict = None):
        """
        Settings combobox

        :param window: main window
        :param id: option id
        :param parent_id: parent option id
        :param option: option data
        """
        super(OptionCombo, self).__init__(window)
        self.window = window
        self.id = id
        self.parent_id = parent_id
        self.option = option
        self.value = None
        self.keys = []
        self.title = ""
        self.real_time = False
        self.combo = NoScrollCombo()
        self.combo.currentIndexChanged.connect(self.on_combo_change)
        self.current_id = None

        # add items
        self.update()

        self.layout = QHBoxLayout()
        self.layout.addWidget(self.combo)
        self.setLayout(self.layout)
        self.fit_to_content()

    def update(self):
        """Prepare items"""
        # init from option data
        if self.option is not None:
            if "label" in self.option and self.option["label"] is not None and self.option["label"] != "":
                self.title = trans(self.option["label"])
            if "keys" in self.option:
                self.keys = self.option["keys"]
            if "value" in self.option:
                self.value = self.option["value"]
                self.current_id = self.value
            if "real_time" in self.option:
                self.real_time = self.option["real_time"]

        # add items
        if type(self.keys) is list:
            for item in self.keys:
                if type(item) is dict:
                    for key, value in item.items():
                        self.combo.addItem(value, key)
                else:
                    self.combo.addItem(item, item)

    def set_value(self, value):
        """
        Set value

        :param value: value
        """
        index = self.combo.findData(value)
        if index != -1:
            self.combo.setCurrentIndex(index)

    def get_value(self):
        """
        Get value

        :return: value
        """
        return self.current_id

    def set_keys(self, keys):
        """
        Set keys

        :param keys: keys
        """
        self.keys = keys
        self.option["keys"] = keys
        self.combo.clear()
        self.update()

    def on_combo_change(self, index):
        """
        On combo change

        :param index: combo index
        :return:
        """
        self.current_id = self.combo.itemData(index)
        self.window.controller.config.combo.on_update(self.parent_id, self.id, self.option, self.current_id)

    def fit_to_content(self):
        """Fit to content"""
        self.combo.setSizeAdjustPolicy(QComboBox.AdjustToContents)




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.05.02 19:00:00                  #
# ================================================== #

from PySide6.QtCore import QModelIndex

from pygpt_net.core.dispatcher import Event
from pygpt_net.core.access.events import AppEvent
from pygpt_net.item.ctx import CtxItem

from .common import Common
from .summarizer import Summarizer
from .extra import Extra

from pygpt_net.utils import trans


class Ctx:
    def __init__(self, window=None):
        """
        Context controller

        :param window: Window instance
        """
        self.window = window
        self.common = Common(window)
        self.summarizer = Summarizer(window)
        self.extra = Extra(window)

        # current edit IDs
        self.edit_meta_id = None
        self.edit_item_id = None

        # current group ID
        self.group_id = None

    def setup(self):
        """Setup ctx"""
        self.common.restore_display_filter()  # load filters first

        # load ctx list
        self.window.core.ctx.load_meta()

        # if no context yet then create one
        if self.window.core.ctx.count_meta() == 0:
            self.new()
        else:
            # get last ctx from config
            id = self.window.core.config.get('ctx')
            if id is not None and self.window.core.ctx.has(id):
                self.window.core.ctx.current = id
            else:
                # if no ctx then get first ctx
                self.window.core.ctx.current = self.window.core.ctx.get_first()

        # load selected ctx
        self.load(self.window.core.ctx.current)

        # restore search string if exists
        if self.window.core.config.has("ctx.search.string"):
            string = self.window.core.config.get("ctx.search.string")
            if string is not None and string != "":
                self.window.ui.nodes['ctx.search'].setText(string)
                self.search_string_change(string)
                # check if current selected ctx is still valid
                if self.window.core.ctx.current is not None:
                    if not self.window.core.ctx.has(self.window.core.ctx.current):
                        self.search_string_clear()
                        # ^ clear search and reload ctx list to prevent creating new ctx

        self.window.ui.nodes['ctx.list'].collapseAll()  # collapse all items at start
        self.restore_expanded_groups()  # restore expanded groups

    def update(self, reload: bool = True, all: bool = True, select: bool = True):
        """
        Update ctx list

        :param reload: reload ctx list items
        :param all: update all
        :param select: select current ctx
        """
        # reload ctx list items
        if reload:
            self.update_list(True)

        # select current ctx on list
        if select:
            self.select_by_current()

        # update all
        if all:
            self.window.controller.ui.update()

        # append ctx and thread id (assistants API) to config
        id = self.window.core.ctx.current
        if id is not None:
            self.window.core.config.set('ctx', id)
            self.window.core.config.set('assistant_thread', self.window.core.ctx.thread)
            self.window.core.config.save()

        # update calendar ctx list
        self.window.controller.calendar.update(all=False)

    def select(self, id: int):
        """
        Select ctx by id

        :param id: context id
        """
        prev_id = self.window.core.ctx.current
        self.window.core.ctx.current = id
        if prev_id != id:
            self.load(id)
            self.window.core.dispatcher.dispatch(AppEvent(AppEvent.CTX_SELECTED))  # app event
        else:
            # only update current group if defined
            meta = self.window.core.ctx.get_meta_by_id(id)
            if meta is not None:
                self.set_group(meta.group_id)
        self.common.focus_chat()

    def select_by_idx(self, idx: int):
        """
        Select ctx by index

        :param idx: context index
        """
        self.select_by_id(self.window.core.ctx.get_id_by_idx(idx))

    def select_by_id(self, id: int):
        """
        Select ctx by index

        :param id: context index
        """
        # lock if generating response is in progress
        if self.context_change_locked():
            return

        self.select(id)
        event = Event(Event.CTX_SELECT, {
            'value': id,
        })
        self.window.core.dispatcher.dispatch(event)

    def select_by_current(self, focus: bool = False):
        """
        Select ctx by current

        :param focus: focus chat
        """
        id = self.window.core.ctx.current
        meta = self.window.core.ctx.get_meta()
        if id in meta:
            self.select_index_by_id(id)
        if focus:
            self.update()

    def unselect(self):
        """Unselect ctx"""
        self.set_group(None)
        self.window.ui.nodes['ctx.list'].clearSelection()

    def set_group(self, group_id: int = None):
        """
        Set current selected group

        :param group_id: group ID
        """
        self.group_id = group_id

    def search_focus_in(self):
        """Search focus handler"""
        pass
        # self.select_by_current()

    def new_ungrouped(self):
        """Create new ungrouped ctx"""
        self.group_id = None
        self.new()

    def new(self, force: bool = False, group_id: int = None):
        """
        Create new ctx

        :param force: force context creation
        :param group_id: group ID
        """
        # lock if generating response is in progress
        if not force and self.context_change_locked():
            return

        # use currently selected group if not defined
        if group_id is None:
            if self.group_id is not None and self.group_id > 0:
                group_id = self.group_id
        else:
            self.group_id = group_id

        # check if group exists
        if group_id is not None and not self.window.core.ctx.has_group(group_id):
            group_id = None
            self.group_id = None

        self.window.core.ctx.new(group_id)
        self.window.core.config.set('assistant_thread', None)  # reset assistant thread id
        self.update()

        # reset appended data
        self.window.controller.chat.render.reset()
        self.window.controller.chat.render.clear_output()

        if not force:  # only if real click on new context button
            self.window.controller.chat.common.unlock_input()
            self.window.ui.nodes['input'].setFocus()  # set focus to input

        # update context label
        mode = self.window.core.ctx.mode
        assistant_id = None
        if mode == 'assistant':
            assistant_id = self.window.core.config.get('assistant')
            self.window.controller.assistant.files.update()  # always update assistant files

        self.common.update_label(mode, assistant_id)
        self.common.focus_chat()

        # app event
        self.window.core.dispatcher.dispatch(AppEvent(AppEvent.CTX_CREATED))

    def add(self, ctx: CtxItem):
        """
        Add ctx item (CtxItem object)

        :param ctx: CtxItem
        """
        self.window.core.ctx.add(ctx)
        self.update()

    def prev(self):
        """Select previous ctx"""
        id = self.window.core.ctx.get_prev()
        if id is not None:
            self.select(id)

    def next(self):
        """Select next ctx"""
        id = self.window.core.ctx.get_next()
        if id is not None:
            self.select(id)

    def last(self):
        """Select last (newest) ctx"""
        id = self.window.core.ctx.get_last_meta()
        if id is not None:
            self.select(id)

    def update_list(self, reload: bool = False):
        """
        Reload current ctx list

        :param reload: reload ctx list items
        """
        self.window.ui.contexts.ctx_list.update(
            'ctx.list',
            self.window.core.ctx.get_meta(reload),
        )

    def refresh(self, restore_model: bool = True):
        """
        Refresh context

        :param restore_model: restore model
        """
        self.load(
            self.window.core.ctx.current,
            restore_model,
        )

    def refresh_output(self):
        """Refresh output"""
        # append ctx to output
        self.window.controller.chat.render.append_context(
            self.window.core.ctx.items,
            clear=True,
        )

    def load(self, id: int, restore_model: bool = True):
        """
        Load ctx data

        :param id: context ID
        :param restore_model: restore model if defined in ctx
        """
        # select ctx by id
        self.window.core.ctx.thread = None  # reset thread id
        self.window.core.ctx.select(id, restore_model=restore_model)
        meta = self.window.core.ctx.get_meta_by_id(id)
        if meta is not None:
            self.set_group(meta.group_id)  # set current group if defined

        # reset appended data
        self.window.controller.chat.render.reset()

        # get current settings stored in ctx
        thread = self.window.core.ctx.thread
        mode = self.window.core.ctx.mode
        model = self.window.core.ctx.model
        assistant_id = self.window.core.ctx.assistant
        preset = self.window.core.ctx.preset

        # restore thread from ctx
        self.window.core.config.set('assistant_thread', thread)

        # clear before output and append ctx to output
        self.refresh_output()

        # switch mode to ctx mode
        if mode is not None:
            self.window.controller.mode.set(mode)  # preset reset here

            # switch preset to ctx preset
            if preset is not None:
                self.window.controller.presets.set(mode, preset)
                self.window.controller.presets.refresh()  # update presets only

            # if ctx mode == assistant then switch assistant to ctx assistant
            if mode == 'assistant':
                # if assistant defined then select it
                if assistant_id is not None:
                    self.window.controller.assistant.select_by_id(assistant_id)
                else:
                    # if empty ctx assistant then get assistant from current selected
                    assistant_id = self.window.core.config.get('assistant')
                self.window.controller.assistant.files.update()  # always update assistant files

            # switch model to ctx model if model is defined in ctx and model is available for this mode
            if model is not None and self.window.core.models.has_model(mode, model):
                self.window.controller.model.set(mode, model)

        # reload ctx list and select current ctx on list, without reloading all
        self.update(reload=False, all=True)

        # update current ctx label in UI
        self.common.update_label(mode, assistant_id)

    def update_ctx(self):
        """Update current ctx mode if allowed"""
        mode = self.window.core.config.get('mode')

        id = None
        # update ctx mode only if current ctx is allowed for this mode
        if self.window.core.ctx.is_allowed_for_mode(mode, False):  # do not check assistant match
            self.window.core.ctx.update()

            # update current context label
            if mode == 'assistant':
                if self.window.core.ctx.assistant is not None:
                    # get assistant id from ctx if defined in ctx
                    id = self.window.core.ctx.assistant
                else:
                    # or get assistant id from current selected assistant
                    id = self.window.core.config.get('assistant')

        # update ctx label
        self.common.update_label(mode, id)

    def delete(self, id: int, force: int = False):
        """
        Delete ctx by idx

        :param id: context id
        :param force: force delete
        """
        if not force:
            self.window.ui.dialogs.confirm(
                type='ctx.delete',
                id=id,
                msg=trans('ctx.delete.confirm'),
            )
            return

        # delete data from indexes if exists
        try:
            self.delete_meta_from_idx(id)
        except Exception as e:
            self.window.core.debug.log(e)
            print("Error deleting ctx data from indexes", e)

        # delete ctx items from db
        items = self.window.core.ctx.all()
        self.window.core.history.remove_items(items)  # remove txt history items
        self.window.core.ctx.remove(id)  # remove ctx from db

        # reset current if current ctx deleted
        if self.window.core.ctx.current == id:
            self.window.core.ctx.current = None
            self.window.controller.chat.render.clear_output()
        self.update()

    def delete_meta_from_idx(self, id: int):
        """
        Delete meta from indexes

        :param id: ctx meta id
        """
        meta = self.window.core.ctx.get_meta_by_id(id)
        if meta is None:
            return

        # check if ctx is indexed
        if meta.indexed is not None and meta.indexed > 0:
            for store_id in list(meta.indexes):
                for idx in list(meta.indexes[store_id]):
                    self.window.core.ctx.idx.remove_meta_from_indexed(store_id, id, idx)

    def delete_item(self, id: int, force: bool = False):
        """
        Delete ctx item by id

        :param id: ctx item id
        :param force: force delete
        """
        if not force:
            self.window.ui.dialogs.confirm(
                type='ctx.delete_item',
                id=id,
                msg=trans('ctx.delete.item.confirm'),
            )
            return
        self.window.core.ctx.remove_item(id)
        self.refresh()
        self.update()

    def delete_history(self, force: bool = False):
        """
        Delete all ctx / truncate

        :param force: force delete
        """
        if not force:
            self.window.ui.dialogs.confirm(
                type='ctx.delete_all',
                id='',
                msg=trans('ctx.delete.all.confirm'),
            )
            return

        # truncate index db if exists
        try:
            self.window.core.idx.ctx.truncate()
        except Exception as e:
            self.window.core.debug.log(e)
            print("Error truncating ctx index db", e)

        # truncate ctx and history
        self.group_id = None
        self.unselect()
        self.window.core.ctx.truncate()
        self.window.core.history.truncate()
        self.update()
        self.new()

    def delete_history_groups(self, force: bool = False):
        """
        Delete all ctx / truncate

        :param force: force delete
        """
        if not force:
            self.window.ui.dialogs.confirm(
                type='ctx.delete_all_groups',
                id='',
                msg=trans('ctx.delete.all.confirm'),
            )
            return

        # truncate index db if exists
        try:
            self.window.core.idx.ctx.truncate()
        except Exception as e:
            self.window.core.debug.log(e)
            print("Error truncating ctx index db", e)

        # truncate ctx and history
        self.group_id = None
        self.unselect()
        self.window.core.ctx.truncate()
        self.window.core.history.truncate()
        self.window.core.ctx.truncate_groups()
        self.update()
        self.new()

    def new_if_empty(self):
        """Create new context if empty"""
        self.group_id = None
        if self.window.core.ctx.count_meta() == 0:
            self.new()

    def rename(self, id: int):
        """
        Ctx name rename by id (show dialog)

        :param id: context id
        """
        meta = self.window.core.ctx.get_meta_by_id(id)
        self.window.ui.dialog['rename'].id = 'ctx'
        self.window.ui.dialog['rename'].input.setText(meta.name)
        self.window.ui.dialog['rename'].current = id
        self.window.ui.dialog['rename'].show()
        self.update()

    def set_important(self, id: int, value: bool = True):
        """
        Set as important

        :param id: context idx
        :param value: important value
        """
        meta = self.window.core.ctx.get_meta_by_id(id)
        if meta is not None:
            meta.important = value
            self.window.core.ctx.save(id)
            self.update()

    def is_important(self, idx: int) -> bool:
        """
        Check if ctx is important

        :param idx: context idx
        :return: True if important
        """
        id = self.window.core.ctx.get_id_by_idx(idx)
        meta = self.window.core.ctx.get_meta_by_id(id)
        if meta is not None:
            return meta.important
        return False

    def set_label(self, id: int, label_id: int):
        """
        Set color label for ctx by idx

        :param id: context idx
        :param label_id: label id
        """
        meta = self.window.core.ctx.get_meta_by_id(id)
        if meta is not None:
            meta.label = label_id
            self.window.core.ctx.save(id)
            self.update()

    def update_name(
            self,
            id: int,
            name: str,
            close: bool = True,
            refresh: bool = True
    ):
        """
        Update ctx name

        :param id: context id
        :param name: context name
        :param close: close rename dialog
        :param refresh: refresh ctx list
        """
        if id not in self.window.core.ctx.get_meta():
            return
        self.window.core.ctx.meta[id].name = name
        self.window.core.ctx.set_initialized()
        self.window.core.ctx.save(id)
        if close:
            self.window.ui.dialog['rename'].close()

        if refresh:
            self.update()
        else:
            self.update(True, False)

    def update_name_current(self, name: str):
        """
        Update current ctx name

        :param name: context name
        """
        self.update_name(self.window.core.ctx.current, name)

    def handle_allowed(self, mode: str) -> bool:
        """
        Check if append to current ctx is allowed for this mode, if not then switch to new context

        :param mode: mode name
        :return: True if allowed
        """
        if not self.window.core.ctx.is_allowed_for_mode(mode):
            self.new(True)  # force new context
            return False
        return True

    def selection_change(self):
        """Select ctx on list change"""
        # TODO: implement this
        # idx = self.window.ui.nodes['ctx.list'].currentIndex().row()
        # self.select(idx)
        self.window.ui.nodes['ctx.list'].lockSelection()

    def search_string_change(self, text: str):
        """
        Search string changed handler

        :param text: search string
        """
        self.window.core.ctx.clear_tmp_meta()
        self.window.core.ctx.search_string = text
        self.window.core.config.set('ctx.search.string', text)
        self.update(reload=True, all=False)

    def search_string_clear(self):
        """Search string clear"""
        self.window.ui.nodes['ctx.search'].clear()
        self.search_string_change("")  # clear search

    def append_search_string(self, text: str):
        """
        Append search string to input and make search

        :param text: search string
        """
        self.window.ui.nodes['ctx.search'].setText(text)
        self.search_string_change(text)  # make search

    def label_filters_changed(self, labels: list):
        """
        Filters labels change

        :param labels: list of labels
        """
        self.window.core.ctx.clear_tmp_meta()
        self.window.core.ctx.filters_labels = labels
        self.window.core.config.set('ctx.records.filter.labels', labels)
        self.update(reload=True, all=False)

    def prepare_name(self, ctx: CtxItem):
        """
        Handle context name (summarize first input and output)

        :param ctx: CtxItem
        """
        # if ctx is not initialized yet then summarize
        if not self.window.core.ctx.is_initialized():
            self.summarizer.summarize(
                self.window.core.ctx.current,
                ctx,
            )

    def context_change_locked(self) -> bool:
        """
        Check if ctx change is locked

        :return: True if locked
        """
        return self.window.controller.chat.input.generating

    def select_index_by_id(self, id):
        """
        Select item by ID on context list

        :param id: ctx meta ID
        """
        index = self.get_child_index_by_id(id)
        self.window.ui.nodes['ctx.list'].unlocked = True  # tmp allow change if locked (enable)
        self.window.ui.nodes['ctx.list'].setCurrentIndex(index)
        self.window.ui.nodes['ctx.list'].unlocked = False  # tmp allow change if locked (disable)

    def find_index_by_id(self, item, id):
        """
        Return index of item with given ID, searching recursively through the model.

        :param item: QStandardItem
        :param id: int
        :return: QModelIndex
        """
        if hasattr(item, 'id') and item.id == id:
            return item.index()
        for row in range(item.rowCount()):
            found_index = self.find_index_by_id(item.child(row), id)
            if found_index.isValid():
                return found_index
        return QModelIndex()

    def find_parent_index_by_id(self, item, id):
        """
        Return index of item with given ID, searching recursively through the model.

        :param item: QStandardItem
        :param id: int
        :return: QModelIndex
        """
        if hasattr(item, 'id') and hasattr(item, 'isFolder') and item.isFolder and item.id == id:
            return item.index()
        for row in range(item.rowCount()):
            found_index = self.find_parent_index_by_id(item.child(row), id)
            if found_index.isValid():
                return found_index
        return QModelIndex()

    def get_parent_index_by_id(self, id):
        """
        Return QModelIndex of parent item based on its ID.

        :param id: int
        :return: QModelIndex
        """
        model = self.window.ui.models['ctx.list']
        root = model.invisibleRootItem()
        return self.find_parent_index_by_id(root, id)

    def get_children_index_by_id(self, parent_id, child_id):
        """
        Return QModelIndex of child item based on its ID and parent ID.

        :param parent_id: int
        :param child_id: int
        :return: QModelIndex
        """
        model = self.window.ui.models['ctx.list']
        parent_index = self.get_parent_index_by_id(parent_id)
        if not parent_index.isValid():
            # no parent found
            return QModelIndex()

        parent_item = model.itemFromIndex(parent_index)
        return self.find_index_by_id(parent_item, child_id)

    def find_child_index_by_id(self, root_item, child_id):
        """
        Find and return QModelIndex of child based on its ID, recursively searching through the model.

        :param root_item: QStandardItem
        :param child_id: int
        :return: QModelIndex
        """
        for row in range(root_item.rowCount()):
            item = root_item.child(row)
            if hasattr(item, 'id') and hasattr(item, 'isFolder') and not item.isFolder and item.id == child_id:
                return item.index()
            child_index = self.find_child_index_by_id(item, child_id)
            if child_index.isValid():
                return child_index
        return QModelIndex()

    def get_child_index_by_id(self, child_id):
        """
        Return QModelIndex of child item based on its ID.

        :param child_id: int
        :return: QModelIndex
        """
        model = self.window.ui.models['ctx.list']
        root_item = model.invisibleRootItem()
        return self.find_child_index_by_id(root_item, child_id)

    def store_expanded_groups(self):
        """
        Store expanded groups in ctx list
        """
        expanded = []
        for group_id in self.window.ui.nodes['ctx.list'].expanded_items:
            expanded.append(group_id)
        self.window.core.config.set('ctx.list.expanded', expanded)
        self.window.core.config.save()

    def restore_expanded_groups(self):
        """
        Restore expanded groups in ctx list
        """
        expanded = self.window.core.config.get('ctx.list.expanded')
        if expanded is not None:
            for group_id in expanded:
                self.window.ui.nodes['ctx.list'].expand_group(group_id)

    def save_all(self):
        """Save visible ctx list items"""
        self.store_expanded_groups()

    def move_to_group(self, meta_id, group_id, update: bool = True):
        """
        Move ctx to group

        :param meta_id: int
        :param group_id: int
        :param update: update ctx list
        """
        self.window.core.ctx.update_meta_group_id(meta_id, group_id)
        self.group_id = group_id
        if update:
            self.update()

    def remove_from_group(self, meta_id):
        """
        Remove ctx from group

        :param meta_id: int
        """
        self.window.core.ctx.update_meta_group_id(meta_id, None)
        self.group_id = None
        self.update()

    def new_group(self, meta_id = None):
        """
        Open new group dialog

        :param meta_id: int
        """
        self.window.ui.dialog['create'].id = 'ctx.group'
        self.window.ui.dialog['create'].input.setText("")
        self.window.ui.dialog['create'].current = meta_id
        self.window.ui.dialog['create'].show()
        self.window.ui.dialog['create'].input.setFocus()

    def create_group(self, name: str = None, meta_id = None):
        """
        Make directory

        :param name: name of directory
        :param meta_id: int
        """
        if name is None:
            self.window.update_status(
                "[ERROR] Name is empty."
            )
            return
        group = self.window.core.ctx.make_group(name)
        id = self.window.core.ctx.insert_group(group)
        if id is not None:
            if meta_id is not None:
                self.move_to_group(meta_id, id, update=False)
            self.update()
            self.window.update_status(
                "[INFO] Group '{}' created.".format(name)
            )
            self.window.ui.dialog['create'].close()

            # select new group
            self.select_group(id)
            self.group_id = id

    def rename_group(self, id: int, force: bool = False):
        """
        Rename group

        :param id: group ID
        :param force: force rename
        """
        if not force:
            group = self.window.core.ctx.get_group_by_id(id)
            if group is None:
                return
            self.window.ui.dialog['rename'].id = 'ctx.group'
            self.window.ui.dialog['rename'].input.setText(group.name)
            self.window.ui.dialog['rename'].current = id
            self.window.ui.dialog['rename'].show()

    def update_group_name(self, id: int, name: str, close: bool = True):
        """
        Update group name

        :param id: group ID
        :param name: group name
        :param close: close rename dialog
        """
        group = self.window.core.ctx.get_group_by_id(id)
        if group is not None:
            group.name = name
            self.window.core.ctx.update_group(group)
            if close:
                self.window.ui.dialog['rename'].close()
            self.update(True, False, False)
            self.select_group(id)

    def get_group_name(self, id: int) -> str:
        """
        Get group name by ID

        :param id: group ID
        :return: group name
        """
        group = self.window.core.ctx.get_group_by_id(id)
        if group is not None:
            return group.name
        return ""

    def select_group(self, id: int):
        """
        Select group

        :param id: group ID
        """
        self.group_id = id
        index = self.get_parent_index_by_id(id)
        self.window.ui.nodes['ctx.list'].unlocked = True  # tmp allow change if locked (enable)
        self.window.ui.nodes['ctx.list'].setCurrentIndex(index)
        self.window.ui.nodes['ctx.list'].unlocked = False  # tmp allow change if locked (disable)

    def delete_group(self, id: int, force: bool = False):
        """
        Delete group only

        :param id: group ID
        :param force: force delete
        """
        if not force:
            self.window.ui.dialogs.confirm(
                type='ctx.group.delete',
                id=id,
                msg=trans('confirm.ctx.delete')
            )
            return
        group = self.window.core.ctx.get_group_by_id(id)
        if group is not None:
            self.window.core.ctx.remove_group(group, all=False)
            if self.group_id == id:
                self.group_id = None
            self.update()

    def delete_group_all(self, id: int, force: bool = False):
        """
        Delete group with all items

        :param id: group ID
        :param force: force delete
        """
        if not force:
            self.window.ui.dialogs.confirm(
                type='ctx.group.delete.all',
                id=id,
                msg=trans('confirm.ctx.delete.all')
            )
            return
        group = self.window.core.ctx.get_group_by_id(id)
        if group is not None:
            self.window.core.ctx.remove_group(group, all=True)
            if self.group_id == id:
                self.group_id = None
            self.update()

    def reload(self):
        """Reload ctx"""
        self.window.core.ctx.reset()
        self.setup()
        self.update()
        self.refresh()

    def reload_after(self):
        """After reload"""
        self.new_if_empty()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.04.30 16:00:00                  #
# ================================================== #

import os

from PySide6.QtCore import QObject, Signal, QRunnable, Slot


class Importer:
    def __init__(self, window=None):
        """
        Importer core

        :param window: Window instance
        """
        self.window = window

    @Slot(str, object)
    def handle_error(self, mode: str, err: any):
        """
        Handle thread error signal

        :param mode: mode
        :param err: error message
        """
        if mode == "assistants":
            self.window.controller.assistant.batch.handle_imported_assistants_failed(err)
        elif mode == "import_files":
            self.window.controller.assistant.batch.handle_imported_files_failed(err)
        elif mode == "truncate_files":
            self.window.controller.assistant.batch.handle_truncated_files_failed(err)
        elif mode == "upload_files":
            self.window.controller.assistant.batch.handle_uploaded_files_failed(err)
        elif mode in "vector_stores":
            self.window.controller.assistant.batch.handle_imported_stores_failed(err)
        elif mode in "truncate_vector_stores":
            self.window.controller.assistant.batch.handle_truncated_stores_failed(err)
        elif mode in "refresh_vector_stores":
            self.window.controller.assistant.batch.handle_refreshed_stores_failed(err)

    @Slot(str, str, int)
    def handle_finished(self, mode: str, store_id: str = None, num: int = 0):
        """
        Handle thread finished signal

        :param mode: mode
        :param store_id: store ID
        :param num: number of affected items
        """
        if mode == "assistants":
            self.window.controller.assistant.batch.handle_imported_assistants(num)
        elif mode == "import_files":
            self.window.controller.assistant.batch.handle_imported_files(num)
        elif mode == "truncate_files":
            self.window.controller.assistant.batch.handle_truncated_files(store_id, num)
        elif mode == "upload_files":
            self.window.controller.assistant.batch.handle_uploaded_files(num)
        elif mode == "vector_stores":
            self.window.controller.assistant.batch.handle_imported_stores(num)
        elif mode == "truncate_vector_stores":
            self.window.controller.assistant.batch.handle_truncated_stores(num)
        elif mode == "refresh_vector_stores":
            self.window.controller.assistant.batch.handle_refreshed_stores(num)

    @Slot(str, str)
    def handle_status(self, mode: str, msg: str):
        """
        Handle thread status change signal

        :param mode: mode
        :param msg: message
        """
        self.window.controller.assistant.batch.handle_status_change(mode, msg)

    @Slot(str, str)
    def handle_log(self, mode: str, msg: str):
        """
        Handle thread log message signal

        :param mode: mode
        :param msg: message
        """
        self.window.controller.assistant.threads.log(mode + ": " + msg)

    def import_assistants(self):
        """Import assistants"""
        worker = ImportWorker()
        worker.window = self.window
        worker.mode = "assistants"
        self.connect_signals(worker)
        self.window.threadpool.start(worker)

    def import_vector_stores(self):
        """Import vector stores"""
        worker = ImportWorker()
        worker.window = self.window
        worker.mode = "vector_stores"
        self.connect_signals(worker)
        self.window.threadpool.start(worker)

    def truncate_vector_stores(self):
        """Truncate vector stores"""
        worker = ImportWorker()
        worker.window = self.window
        worker.mode = "truncate_vector_stores"
        self.connect_signals(worker)
        self.window.threadpool.start(worker)

    def truncate_files(self, store_id: str = None):
        """
        Truncate files

        :param store_id: store ID
        """
        worker = ImportWorker()
        worker.window = self.window
        worker.mode = "truncate_files"
        worker.store_id = store_id
        self.connect_signals(worker)
        self.window.threadpool.start(worker)

    def upload_files(self, store_id: str, files: list = None):
        """
        Upload files

        :param store_id: store ID
        :param files: files
        """
        worker = ImportWorker()
        worker.window = self.window
        worker.mode = "upload_files"
        worker.store_id = store_id
        worker.files = files
        self.connect_signals(worker)
        self.window.threadpool.start(worker)

    def refresh_vector_stores(self):
        """Refresh vector stores"""
        worker = ImportWorker()
        worker.window = self.window
        worker.mode = "refresh_vector_stores"
        self.connect_signals(worker)
        self.window.threadpool.start(worker)

    def import_files(self, store_id: str = None):
        """
        Import assistant files

        :param store_id: store ID
        """
        worker = ImportWorker()
        worker.window = self.window
        worker.mode = "import_files"
        worker.store_id = store_id
        self.connect_signals(worker)
        self.window.threadpool.start(worker)

    def connect_signals(self, worker):
        """
        Connect signals

        :param worker: worker instance
        """
        worker.signals.finished.connect(self.handle_finished)
        worker.signals.error.connect(self.handle_error)
        worker.signals.status.connect(self.handle_status)
        worker.signals.log.connect(self.handle_log)


class ImportWorkerSignals(QObject):
    """Import worker signals"""
    status = Signal(str, str)  # mode, message
    finished = Signal(str, str, int)  # mode, store_id, num
    error = Signal(str, object)  # mode, error
    log = Signal(str, str)  # mode, message


class ImportWorker(QRunnable):
    """Import worker"""
    def __init__(self, *args, **kwargs):
        super(ImportWorker, self).__init__()
        self.signals = ImportWorkerSignals()
        self.window = None
        self.mode = "assistants"
        self.assistant = None
        self.store_id = None
        self.files = []

    @Slot()
    def run(self):
        """Importer thread"""
        try:
            # import data
            if self.mode == "assistants":
                self.import_assistants()
            elif self.mode == "vector_stores":
                if self.import_vector_stores():
                    self.import_files()
            elif self.mode == "truncate_vector_stores":
                self.truncate_vector_stores()
            elif self.mode == "refresh_vector_stores":
                self.refresh_vector_stores()
            elif self.mode == "truncate_files":
                self.truncate_files()
            elif self.mode == "import_files":
                self.import_files()
            elif self.mode == "upload_files":
                self.upload_files()
        except Exception as e:
            self.signals.error.emit(self.mode, e)

    def import_assistants(self, silent: bool = False) -> bool:
        """
        Import assistants from API

        :param silent: silent mode (no signals)
        :return: result
        """
        try:
            # import assistants
            self.log("Importing assistants...")
            self.window.core.assistants.clear()
            items = self.window.core.assistants.get_all()
            self.window.core.gpt.assistants.import_all(items, callback=self.callback)
            self.window.core.assistants.items = items
            self.window.core.assistants.save()

            # import vector stores
            self.import_vector_stores(True)

            # import files
            self.import_files(True)

            if not silent:
                self.signals.finished.emit("assistants", self.store_id, len(items))
            return True
        except Exception as e:
            self.log("API error: {}".format(e))
            self.signals.error.emit("assistants", e)
            return False

    def import_vector_stores(self, silent: bool = False) -> bool:
        """
        Import vector stores from API

        :param silent: silent mode (no signals)
        :return: result
        """
        try:
            self.log("Importing vector stores...")
            self.window.core.assistants.store.clear()
            items = {}
            self.window.core.gpt.store.import_stores(items, callback=self.callback)
            self.window.core.assistants.store.import_items(items)
            if not silent:
                self.signals.finished.emit("vector_stores", self.store_id, len(items))
            return True
        except Exception as e:
            self.log("API error: {}".format(e))
            self.signals.error.emit("vector_stores", e)
            return False

    def truncate_vector_stores(self, silent: bool = False) -> bool:
        """
        Truncate all vector stores in API

        :param silent: silent mode (no signals)
        :return: result
        """
        try:
            self.log("Truncating stores...")
            num = self.window.core.gpt.store.remove_all(callback=self.callback)
            self.window.core.assistants.store.items = {}
            self.window.core.assistants.store.save()
            if not silent:
                self.signals.finished.emit("truncate_vector_stores", self.store_id, num)
            return True
        except Exception as e:
            self.log("API error: {}".format(e))
            self.signals.error.emit("truncate_vector_stores", e)
            return False

    def refresh_vector_stores(self, silent: bool = False) -> bool:
        """
        Refresh all vector stores in API

        :param silent: silent mode (no signals)
        :return: result
        """
        try:
            self.log("Refreshing stores...")
            num = 0
            stores = self.window.core.assistants.store.items
            for id in stores:
                store = stores[id]
                try:
                    self.window.controller.assistant.store.refresh_store(store, update=False)
                    num += 1
                except Exception as e:
                    self.log("Failed to refresh store: {}".format(id))
                    self.window.core.debug.log(e)
            if not silent:
                self.signals.finished.emit("refresh_vector_stores", self.store_id, num)
            return True
        except Exception as e:
            self.log("API error: {}".format(e))
            self.signals.error.emit("refresh_vector_stores", e)
            return False

    def truncate_files(self, silent: bool = False) -> bool:
        """
        Truncate files in API

        :param silent: silent mode (no signals)
        :return: result
        """
        try:
            # if empty store_id, truncate all files, otherwise truncate only store files
            if self.store_id is None:
                self.log("Truncating all files...")
                self.window.core.assistants.files.truncate() # clear all files
                # remove all files in API
                num = self.window.core.gpt.store.remove_files(callback=self.callback)
            else:
                self.log("Truncating files for store: {}".format(self.store_id))
                self.window.core.assistants.files.truncate(self.store_id)  # clear store files, remove from stores / DB
                # remove store files in API
                num = self.window.core.gpt.store.remove_store_files(
                    self.store_id,
                    callback=self.callback,
                )
            if not silent:
                self.signals.finished.emit("truncate_files", self.store_id, num)
            return True
        except Exception as e:
            self.log("API error: {}".format(e))
            self.signals.error.emit("truncate_files", e)
            return False

    def upload_files(self, silent: bool = False) -> bool:
        """
        Upload files to API

        :param silent: silent mode (no signals)
        :return: result
        """
        num = 0
        try:
            self.log("Uploading files...")
            for file in self.files:
                try:
                    file_id = self.window.core.gpt.store.upload(file)
                    if file_id is not None:
                        stored_file = self.window.core.gpt.store.add_file(
                            self.store_id,
                            file_id,
                        )
                        if stored_file is not None:
                            data = self.window.core.gpt.store.get_file(file_id)
                            self.window.core.assistants.files.insert(self.store_id, data)  # insert to DB
                            msg = "Uploaded file: {}/{}".format((num + 1), len(self.files))
                            self.signals.status.emit("upload_files", msg)
                            self.log(msg)
                            num = num + 1
                except Exception as e:
                    self.window.core.debug.log(e)
                    self.signals.status.emit("upload_files", "Failed to upload file: {}".format(os.path.basename(file)))
            if not silent:
                self.signals.finished.emit("upload_files", self.store_id, num)
            return True
        except Exception as e:
            self.log("API error: {}".format(e))
            self.signals.error.emit("upload_files", e)
            return False

    def import_files(self, silent: bool = False) -> bool:
        """
        Import assistant files from API

        :param silent: silent mode (no signals)
        :return: result
        """
        try:
            if self.store_id is None:
                self.log("Importing all files...")
                self.window.core.assistants.files.truncate_local()  # clear local DB (all)
                num = self.window.core.gpt.store.import_stores_files(self.callback)  # import all files
            else:
                self.log("Importing files for store: {}".format(self.store_id))
                self.window.core.assistants.files.truncate_local(self.store_id)  # clear local DB (all)
                items = self.window.core.gpt.store.import_store_files(
                    self.store_id,
                    [],
                    callback=self.callback,
                )  # import store files
                num = len(items)
            if not silent:
                self.signals.finished.emit("import_files", self.store_id, num)
            return True
        except Exception as e:
            self.log("API error: {}".format(e))
            self.signals.error.emit("import_files", e)
        return False

    def callback(self, msg: str):
        """
        Log callback

        :param msg: message
        """
        self.log(msg)

    def log(self, msg: str):
        """
        Log message

        :param msg: message
        """
        self.signals.log.emit(self.mode, msg)
